from __future__ import annotations


import os
import json
import pathlib
import multiprocessing as mp
from pydantic import AliasChoices, Field
from lzo.types import BaseSettings, eproperty, model_validator, field_validator, Literal

from typing import Optional, Dict, Any, List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from ..spec.providers.main import ProviderFileSystemManager

class ConfigMixin(BaseSettings):
    """
    A Config Mixin
    """
    authz_config_dir: Optional[pathlib.Path] = None
    boto_config: Optional[pathlib.Path] = None
    max_workers: Optional[int] = None
    max_pool_connections: Optional[int] = 100

    write_chunking_enabled: Optional[bool] = False
    write_chunking_size: Optional[int] = 1024 * 1024 * 10 # 10 MB
    # if the object size is larger than this, it will use an alternate method
    write_chunking_large_size: Optional[int] = 1024 * 1024 * 50 # 50 MB 

    # Use the transfer manager by default (faster)
    write_chunking_manager_default: Optional[bool] = True 
    
    read_chunking_enabled: Optional[bool] = False
    read_chunking_size: Optional[int] = 1024 * 1024 * 10 # 10 MB
    # if the object size is larger than this, it will use an alternate method
    read_chunking_large_size: Optional[int] = 1024 * 1024 * 50 # 50 MB
    read_chunking_manager_default: Optional[bool] = True

    @field_validator("authz_config_dir", mode = 'before')
    def validate_authz_config_dir(cls, value: Optional[pathlib.Path]) -> pathlib.Path:
        """
        Validates the authz config dir
        """
        return pathlib.Path("~/.auth").expanduser() if value is None else value
    
    @eproperty
    def num_workers(self):
        """
        The number of workers
        """
        return self.max_workers or min(2, round(mp.cpu_count() // 2))

    @eproperty
    def boto_config_path(self) -> pathlib.Path:
        """
        Returns the boto config path
        """
        if self.boto_config is None:
            return pathlib.Path('/root/.boto') if self.in_colab else pathlib.Path("~/.boto").expanduser()
        return self.boto_config

    @eproperty
    def boto_config_exists(self):
        """
        Returns whether the boto config exists
        """
        return self.boto_config_path.exists()

    @eproperty
    def user_home(self) -> pathlib.Path:
        """
        Returns the user home directory
        """
        if self.in_colab:
            return pathlib.Path("/content")
        return pathlib.Path("~").expanduser()
    
    @eproperty
    def in_colab(self) -> bool:
        """
        Checks if the code is running in Google Colab
        """
        from lzl.require import LazyLib
        return LazyLib.is_available('google.colab')
    
    @eproperty
    def provider_fsm(self) -> 'ProviderFileSystemManager':
        """
        Returns the Provider File System Manager
        """
        from ..spec.providers.main import ProviderManager
        return ProviderManager
    
    def set_env(self):
        """
        Sets the environment variables for boto
        """
        if self.boto_config_exists:
            os.environ["BOTO_CONFIG"] = self.boto_config_path.as_posix()
            os.environ["BOTO_PATH"] = self.boto_config_path.as_posix()
    
    def update_config(self, **kwargs):
        """
        Updates the settings
        """
        for k, v in kwargs.items():
            if not hasattr(self, k):  continue
            if isinstance(getattr(self, k), pathlib.Path):
                setattr(self, k, pathlib.Path(v))
            else:
                setattr(self, k, v)

    def update_fs(self, **kwargs):
        """
        Updates the fs config
        """
        pass

    def update_auth(self, update_fs: bool = True, **config):
        """
        Updates the auth
        """
        self.update_config(**config)
        self.set_env()
        if update_fs: self.update_fs(**config)
    
    
    def build_fs_config(self) -> Dict[str, Any]:
        """
        Builds the fsspec config dict
        """
        return {}

    def get_boto_config(self) -> Dict[str, Any]:
        """
        Returns the boto config
        """
        raise NotImplementedError
    

    def get_boto_client_config(self) -> Dict[str, Any]:
        """
        Returns the boto client config
        """
        raise NotImplementedError
    
    class Config:
        case_sensitive = False
        extra = 'allow'

class AWSConfig(ConfigMixin):
    """
    The AWS Config
    """
    aws_access_token: Optional[str] = None
    aws_access_key_id: Optional[str] = Field(None, validation_alias=AliasChoices('AWS_ACCESS_KEY_ID', 'AWS_ACCESS_KEY'))
    aws_secret_access_key: Optional[str] = Field(None, validation_alias=AliasChoices('AWS_SECRET_ACCESS_KEY', 'AWS_SECRET_KEY'))
    aws_region: Optional[str] = "us-east-1"
    aws_s3_endpoint: Optional[str] = None
    set_s3_endpoint: Optional[bool] = False

    s3_use_accelerate_endpoint: Optional[bool] = False
    s3_config: Optional[Union[str, Dict[str, Any]]] = None
    s3_fs_config: Optional[Union[str, Dict[str, Any]]] = None

    @field_validator("s3_config", mode = 'before')
    def validate_s3_config(cls, value: Optional[Union[str, Dict[str, Any]]]) -> Union[str, Dict[str, Any]]:
        """
        Validates the S3 Config
        """
        if value is None: return {}
        if isinstance(value, str):
            value = json.loads(value)
        return value
    
    @field_validator('aws_s3_endpoint', mode='before')
    def validate_aws_s3_endpoint(cls, value: Optional[str]) -> str:
        """
        Validates the AWS S3 Endpoint
        """
        if value is None: return None
        if not value.startswith('http'):
            # Check if there's a port specified
            prefix = 'https'
            if ':' in value:
                # Likely a http endpoint
                if not value.endswith(':443'): 
                    prefix = 'http'
            value = f'{prefix}://{value}'
        return value
    
    @eproperty
    def s3_endpoint(self) -> str:
        """
        The S3 Endpoint
        """
        if self.aws_s3_endpoint: return self.aws_s3_endpoint
        return f'https://s3.{self.aws_region}.amazonaws.com'
    
    @eproperty
    def has_custom_endpoint(self) -> bool:
        """
        Returns whether the endpoint is custom
        """
        return self.aws_s3_endpoint is not None
    
    def set_env(self):
        """
        Sets the environment variables
        """
        super().set_env()
        if self.aws_access_key_id:
            os.environ['AWS_ACCESS_KEY_ID'] = self.aws_access_key_id
        if self.aws_secret_access_key:
            os.environ['AWS_SECRET_ACCESS_KEY'] = self.aws_secret_access_key
        if self.aws_region:
            os.environ['AWS_REGION'] = self.aws_region
        if self.aws_access_token:
            os.environ['AWS_ACCESS_TOKEN'] = self.aws_access_token
        if self.set_s3_endpoint:
            os.environ['S3_ENDPOINT'] = self.s3_endpoint

    def update_fs(self, **kwargs):
        """
        Updates the fs config
        """
        self.provider_fsm.get_accessor('aws', _reset = True)

    def build_fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config = {
            **(self.s3_fs_config or {})
        }
        if self.aws_access_key_id:
            config["key"] = self.aws_access_key_id
            # config['aws_access_key_id'] = self.aws_access_key_id
        if self.aws_secret_access_key:
            config["secret"] = self.aws_secret_access_key
            # config['aws_secret_access_key'] = self.aws_secret_access_key
        if self.aws_access_token:
            config["token"] = self.aws_access_token
        if not (config.get('key') and config.get('secret')) and not self.boto_config_exists:
            config['anon'] = True
        if self.set_s3_endpoint:
            config['client_kwargs'] = {'endpoint_url': self.s3_endpoint, 'region_name': self.aws_region}
        if self.s3_config:
            config['config_kwargs'] = self.s3_config
        return config

    def get_boto_config(self) -> Dict[str, Any]:
        """
        Returns the boto config
        """
        return {
            'region_name': self.aws_region,
            'max_pool_connections': self.max_pool_connections,
            's3': {
                'use_accelerate_endpoint': self.s3_use_accelerate_endpoint,
            }
        }
    
    def get_boto_client_config(self) -> Dict[str, Any]:
        """
        Returns the boto client config
        """
        base = {
            'aws_access_key_id': self.aws_access_key_id,
            'aws_secret_access_key': self.aws_secret_access_key,
        }
        if self.has_custom_endpoint:
            base['endpoint_url'] = self.s3_endpoint
            base['region_name'] = self.aws_region
        return base
        


class GCPConfig(ConfigMixin):
    """
    The GCP Config
    """

    gcp_project: Optional[str] = None
    gcloud_project: Optional[str] = None
    google_cloud_project: Optional[str] = None
    google_application_credentials: Optional[pathlib.Path] = None

    gcs_client_config: Optional[Union[str, Dict[str, Any]]] = None
    gcs_config: Optional[Union[str, Dict[str, Any]]] = None
    gcs_fs_config: Optional[Union[str, Dict[str, Any]]] = None


    @field_validator('gcs_config', 'gcs_client_config', mode='before')
    def validate_gcs_client_config(cls, value: Optional[Union[str, Dict[str, Any]]]) -> Union[str, Dict[str, Any]]:
        """
        Validates the GCS Client Config
        """
        if value is None: return {}
        if isinstance(value, str):
            value = json.loads(value)
        return value
    
    @model_validator(mode = 'after')
    def validate_gcp(self):
        """
        Validates the GCP Config
        """
        if not self.google_application_credentials:
            self.google_application_credentials = self.user_home.joinpath('adc.json')    
        return self
    
    @eproperty
    def adc_exists(self) -> bool:
        """
        Returns whether the ADC exists
        """
        return self.google_application_credentials.exists()
    
    
    @eproperty
    def project(self):
        return self.gcp_project or self.gcloud_project or self.google_cloud_project
    
    def set_env(self):
        """
        Sets the environment variables
        """
        if self.adc_exists:
            os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = self.google_application_credentials.as_posix()
        if self.project:
            os.environ['GOOGLE_CLOUD_PROJECT'] = self.project
    

    def update_fs(self, **kwargs):
        """
        Updates the fs config
        """
        self.provider_fsm.get_accessor('gcs', _reset = True)


    def build_fs_config(self) -> Dict[str, Any]:
        """
        Builds the gcsfs config kwargs
        """
        config = {
            **(self.gcs_fs_config or {})
        }
        if self.adc_exists: config['token'] = self.google_application_credentials.as_posix()
        if self.project: config['project'] = self.project
        if self.gcs_client_config: config['client_config'] = self.gcs_client_config
        if self.gcs_config: config['config_kwargs'] = self.gcs_config
        return config


class MinioConfig(ConfigMixin):
    """
    The Minio Config
    """
    minio_endpoint: Optional[str] = None
    minio_access_key: Optional[str] = Field(None, validation_alias=AliasChoices('MINIO_ACCESS_KEY', 'MINIO_ACCESS_KEY_ID'))
    minio_secret_key: Optional[str] = Field(None, validation_alias=AliasChoices('MINIO_SECRET_KEY', 'MINIO_SECRET_ACCESS_KEY'))
    minio_access_token: Optional[str] = None
    minio_secure: Optional[bool] = True
    minio_region: Optional[str] = 'us-east-1'
    minio_signature_ver: Optional[str] = 's3v4'
    minio_addressing_style: Optional[Literal['virtual', 'path', 'auto']] = 'auto'
    
    minio_config: Optional[Union[str, Dict[str, Any]]] = None
    minio_fs_config: Optional[Union[str, Dict[str, Any]]] = None

    write_chunking_enabled: Optional[bool] = True
    read_chunking_enabled: Optional[bool] = True
    write_chunking_manager_default: Optional[bool] = True
    read_chunking_manager_default: Optional[bool] = True
    

    @field_validator("minio_fs_config",'minio_config',  mode = 'before')
    def validate_minio_fs_config(cls, value: Optional[Union[str, Dict[str, Any]]]) -> Union[str, Dict[str, Any]]:
        """
        Validates the Minio FS Config
        """
        if value is None: return {}
        if isinstance(value, str):
            value = json.loads(value)
        return value
    

    def set_env(self):
        """
        Sets the environment variables
        """
        if self.minio_endpoint:
            os.environ["MINIO_ENDPOINT"] = self.minio_endpoint
        if self.minio_access_key:
            os.environ["MINIO_ACCESS_KEY"] = self.minio_access_key
        if self.minio_secret_key:
            os.environ["MINIO_SECRET_KEY"] = self.minio_secret_key
        if self.minio_secure:
            os.environ["MINIO_SECURE"] = str(self.minio_secure)
        if self.minio_region:
            os.environ["MINIO_REGION"] = self.minio_region
        if self.minio_signature_ver:
            os.environ["MINIO_SIGNATURE_VER"] = self.minio_signature_ver


    def build_fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config: Dict[str, Union[str, Dict[str, Any], Any]] = {
            "client_kwargs": {
                "endpoint_url": self.minio_endpoint,
                "region_name": self.minio_region,
            },
            "config_kwargs": {
                "signature_version": self.minio_signature_ver,
                "s3": {
                    "addressing_style": self.minio_addressing_style,
                }
            },
            **(self.minio_fs_config or {})
        }
        if self.minio_access_key:
            config["key"] = self.minio_access_key
        if self.minio_secret_key:
            config["secret"] = self.minio_secret_key
        if self.minio_access_token:
            config["token"] = self.minio_access_token
        if self.minio_config:
            config["config_kwargs"].update(self.minio_config)
        return config
    
    def update_fs(self, **kwargs):
        """
        Updates the fs config
        """
        self.provider_fsm.get_accessor('minio', _reset = True)


    def get_boto_config(self) -> Dict[str, Any]:
        """
        Returns the boto config
        """
        return {
            'signature_version': self.minio_signature_ver,
            'max_pool_connections': self.max_pool_connections,
            's3': {
                'addressing_style': self.minio_addressing_style,
            }
        }
    
    def get_boto_client_config(self) -> Dict[str, Any]:
        """
        Returns the boto client config
        """
        return {
            'aws_access_key_id': self.minio_access_key,
            'aws_secret_access_key': self.minio_secret_key,
            'region_name': self.minio_region,
            'endpoint_url': self.minio_endpoint,
            
        }
        
        

class S3CompatConfig(ConfigMixin):
    """
    The S3 Compat Config
    """

    s3c_endpoint: Optional[str] = None
    s3c_access_key: Optional[str] = Field(None, validation_alias=AliasChoices('S3C_ACCESS_KEY', 'S3_COMPAT_ACCESS_KEY'))
    s3c_secret_key: Optional[str] = Field(None, validation_alias=AliasChoices('S3C_SECRET_KEY', 'S3_COMPAT_SECRET_KEY'))
    s3c_access_token: Optional[str] = None
    s3c_secure: Optional[bool] = True
    s3c_region: Optional[str] = None
    s3c_addressing_style: Optional[Literal['virtual', 'path', 'auto']] = 'auto'
    s3c_signature_ver: Optional[str] = 's3v4'

    s3c_config: Optional[Union[str, Dict[str, Any]]] = None
    s3c_fs_config: Optional[Union[str, Dict[str, Any]]] = None

    @model_validator(mode = 'before')
    def map_depreciated_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Maps the depreciated values
        """
        fields = list(cls.model_fields.keys())
        for k in fields:
            dep_k = k.replace('s3c', 's3_compat', 1)
            if dep_k in values:
                val = values.pop(dep_k)
                if k not in values:
                    values[k] = val
        
        # Just handle the config stuff too
        if values.get('s3c_config'):
            if isinstance(values['s3c_config'], str):
                values['s3c_config'] = json.loads(values['s3c_config'])
        if values.get('s3c_fs_config'):
            if isinstance(values['s3c_fs_config'], str):
                values['s3c_fs_config'] = json.loads(values['s3c_fs_config'])
        return values
    
    def set_env(self):
        """
        Set environment variables for the S3C client.
        """
        # Handle the old env vars
        if self.s3c_endpoint:
            os.environ["S3_COMPAT_ENDPOINT"] = self.s3c_endpoint
            os.environ['S3C_ENDPOINT'] = self.s3c_endpoint
        if self.s3c_access_key:
            os.environ["S3_COMPAT_ACCESS_KEY"] = self.s3c_access_key
            os.environ['S3C_ACCESS_KEY'] = self.s3c_access_key
        if self.s3c_secret_key:
            os.environ["S3_COMPAT_SECRET_KEY"] = self.s3c_secret_key
            os.environ['S3C_SECRET_KEY'] = self.s3c_secret_key
        if self.s3c_access_token:
            os.environ["S3_COMPAT_ACCESS_TOKEN"] = self.s3c_access_token
            os.environ['S3C_ACCESS_TOKEN'] = self.s3c_access_token
        if self.s3c_secure:
            os.environ["S3_COMPAT_SECURE"] = str(self.s3c_secure)
            os.environ['S3C_SECURE'] = str(self.s3c_secure)
        if self.s3c_region:
            os.environ["S3_COMPAT_REGION"] = self.s3c_region
            os.environ['S3C_REGION'] = self.s3c_region
        if self.s3c_signature_ver:
            os.environ["S3_COMPAT_SIGNATURE_VER"] = self.s3c_signature_ver
            os.environ['S3C_SIGNATURE_VER'] = self.s3c_signature_ver
    
    def build_fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config: Dict[str, Union[str, Dict[str, Any], Any]] = {
            "client_kwargs": {
                "endpoint_url": self.s3c_endpoint,
                "region_name": self.s3c_region,
            },
            "config_kwargs": {
                "signature_version": self.s3c_signature_ver,
                "s3": {
                    "addressing_style": self.s3c_addressing_style,
                }
            },
            **(self.s3c_fs_config or {})
        }
        if self.s3c_access_key:
            config["key"] = self.s3c_access_key
        if self.s3c_secret_key:
            config["secret"] = self.s3c_secret_key
        if self.s3c_access_token:
            config["token"] = self.s3c_access_token
        if self.s3c_config:
            config["config_kwargs"].update(self.s3c_config)
        return config
    

    def update_fs(self, **kwargs):
        """
        Updates the fs config
        """
        self.provider_fsm.get_accessor('s3c', _reset = True)

    
    def get_boto_config(self) -> Dict[str, Any]:
        """
        Returns the boto config
        """
        return {
            'signature_version': self.s3c_signature_ver,
            'max_pool_connections': self.max_pool_connections,
            's3': {
                'addressing_style': self.s3c_addressing_style,
            }
        }
    
    def get_boto_client_config(self) -> Dict[str, Any]:
        """
        Returns the boto client config
        """
        return {
            'aws_access_key_id': self.s3c_access_key,
            'aws_secret_access_key': self.s3c_secret_key,
            'region_name': self.s3c_region,
            'endpoint_url': self.s3c_endpoint,
        }
        

class CloudflareR2Config(ConfigMixin):
    """
    The Cloudflare R2 Config
    """
    r2_account_id: Optional[str] = None
    r2_access_key_id: Optional[str] = Field(None, validation_alias=AliasChoices('R2_ACCESS_KEY_ID', 'R2_ACCESS_KEY'))
    r2_secret_access_key: Optional[str] = Field(None, validation_alias=AliasChoices('R2_SECRET_ACCESS_KEY', 'R2_SECRET_KEY'))
    r2_access_token: Optional[str] = None
    r2_region: Optional[str] = None

    r2_endpoint: Optional[str] = None
    r2_config: Optional[Union[str, Dict[str, Any]]] = None
    r2_signature_ver: Optional[str] = 's3v4'
    r2_fs_config: Optional[Union[str, Dict[str, Any]]] = None

    @model_validator(mode = 'before')
    def build_r2_values(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """
        Builds the r2 values
        """
        if values.get("r2_config") is not None:
            values["r2_config"] = json.loads(values["r2_config"]) if isinstance(values["r2_config"], str) else values["r2_config"]
        if values.get("r2_fs_config") is not None:
            values["r2_fs_config"] = json.loads(values["r2_fs_config"]) if isinstance(values["r2_fs_config"], str) else values["r2_fs_config"]
        if values.get("r2_endpoint") is not None:
            values["r2_endpoint"] = values["r2_endpoint"].rstrip("/")
            if not values["r2_endpoint"].startswith("http"):
                values["r2_endpoint"] =  "https://" + values["r2_endpoint"]
        elif values.get("r2_account_id"):
            values["r2_endpoint"] = f"https://{values['r2_account_id']}.r2.cloudflarestorage.com"
        return values

    def set_env(self):
        """
        Sets the environment variables
        """
        if self.r2_endpoint:
            os.environ["R2_ENDPOINT"] = self.r2_endpoint
        if self.r2_account_id:
            os.environ["R2_ACCOUNT_ID"] = self.r2_account_id
        if self.r2_access_key_id:
            os.environ["R2_ACCESS_KEY_ID"] = self.r2_access_key_id
        if self.r2_secret_access_key:
            os.environ["R2_SECRET_ACCESS_KEY"] = self.r2_secret_access_key
        if self.r2_access_token:
            os.environ["R2_ACCESS_TOKEN"] = str(self.r2_access_token)
    
    def build_fs_config(self) -> Dict[str, Any]:
        """
        Builds the s3fs config dict
        """
        config: Dict[str, Union[str, Dict[str, Any], Any]] = {
            "client_kwargs": {
                "endpoint_url": self.r2_endpoint,
                "region_name": self.r2_region,
            },
            "config_kwargs": {
                "signature_version": self.r2_signature_ver,
            },
            "fixed_upload_size": True,
            **(self.r2_fs_config or {})
        }
        if self.r2_access_key_id:
            config["key"] = self.r2_access_key_id
        if self.r2_secret_access_key:
            config["secret"] = self.r2_secret_access_key
        if self.r2_access_token:
            config["token"] = self.r2_access_token
        if self.r2_config:
            config["config_kwargs"].update(self.r2_config)
        return config


    def update_fs(self, **kwargs):
        """
        Updates the fs config
        """
        self.provider_fsm.get_accessor('r2', _reset = True)


    def get_boto_config(self) -> Dict[str, Any]:
        """
        Returns the boto config
        """
        return {
            'max_pool_connections': self.max_pool_connections,
        }
    
    def get_boto_client_config(self) -> Dict[str, Any]:
        """
        Returns the boto client config
        """
        return {
            'aws_access_key_id': self.r2_access_key_id,
            'aws_secret_access_key': self.r2_secret_access_key,
            'region_name': self.r2_region,
            'endpoint_url': self.r2_endpoint,
        }
        



from __future__ import annotations

"""
Registry of SQL Functions that can be lazily loaded
"""

import abc
import typing as t
from lzo.registry.base import MRegistry


class SQLFunction(abc.ABC):
    """
    Abstract Base Class for SQL Functions
    """

    _rxtra: t.Dict[str, t.Any] = {}

    def __init_subclass__(cls) -> None:

        return super().__init_subclass__()
    

SQLFunctionT = t.TypeVar('SQLFunctionT', bound = SQLFunction)

class SQLFunctionRegistry(MRegistry[SQLFunctionT]):
    """
    Registry of SQL Functions
    """

    # def re
    



import os
import logging
from lzl.logging import get_logger, change_logger_level, null_logger

# to prevent recursive imports, we'll just use os environ here
logger_level: str = os.getenv('LOGGER_LEVEL', 'INFO').upper()
logger = get_logger(logger_level)

logger.set_module_name('lzl.api.openai', 'openai', is_relative = True)

def configure_httpx_logger(level: int = logging.ERROR):
    """
    Helper Method to Disable HTTPX Logger
    """
    httpx_logger = logging.getLogger("httpx")
    httpx_logger.setLevel(level)



from __future__ import annotations

import os
from pydantic import model_validator, Field, PrivateAttr
from lzl.proxied import ProxyObject
from lzo.types import BaseSettings, AppEnv, eproperty
from lzl.logging import logger, null_logger, Logger

# from lazyops.libs.abcs.configs.base import BaseAppSettings
# from lazyops.libs.abcs.configs.types import AppEnv

from lazyops.libs.abcs.types.persistence import TemporaryData
# from lazyops.libs.logging import logger, null_logger
from typing import Dict, Union, Optional, Any, TYPE_CHECKING
from pathlib import Path

if TYPE_CHECKING:
    # from lazyops.utils.logs import Logger
    from hatchet_sdk.loader import ClientConfig

lib_path = Path(__file__).parent

_temp_data: Optional[TemporaryData] = None

def get_temp_data() -> TemporaryData:
    """
    Retrieves the temporary data
    """
    global _temp_data
    if _temp_data is None:
        _temp_data = TemporaryData(lib_path.joinpath('data'))
    return _temp_data

class HatchetSettings(BaseSettings):
    """
    Hatchet Settings
    """
    base_name: Optional[str] = 'lzo'
    api_endpoint: Optional[str] = None
    grpc_endpoint: Optional[str] = None
    endpoints: Dict[str, Dict[str, str]] = Field(default_factory=dict)
    app_env: Optional[AppEnv] = None
    debug_enabled: Optional[bool] = None
    workflow_concurrency_limit: Optional[int] = 250
    config_lib_path: Optional[str] = None

    include_instance_name_in_workflow: Optional[bool] = False

    _extra: Optional[Dict[str, Any]] = PrivateAttr(default_factory=dict)

    @property
    def temp_data(self) -> 'TemporaryData':
        """
        Retrieves the temporary data
        """
        return get_temp_data()

    class Config:
        env_prefix = "HATCHET_"
    

    def configure_client_endpoints(self, config: 'ClientConfig') -> None:
        # sourcery skip: extract-method
        """
        Configure the client endpoints based on the environment variables
        """
        if self.api_endpoint and self.grpc_endpoint:
            config.server_url = self.api_endpoint
            config.host_port = self.grpc_endpoint
        elif self.grpc_endpoint: config.host_port = self.grpc_endpoint
        elif self.endpoints:
            ep_keys = list(self.endpoints.keys())
            from lazyops.libs.abcs.utils.http import validate_website_with_socket
            endpoints: Dict[str, Dict[str, str]] = None
            if any(key in ep_keys for key in {'api', 'grpc'}):
                refs = list(self.endpoints.get('api', {}).keys())
                # This means there is not env config
                if 'cluster' in refs or 'public' in refs:
                    endpoints = self.endpoints
                else:
                    ref = 'prod' if 'prod' in refs else 'production'
                    if self.is_development_env: ref = 'develop' if 'develop' in refs else 'development'
                    endpoints = {
                        'api': self.endpoints['api'][ref],
                        'grpc': self.endpoints.get('grpc', {}).get(ref),
                    }
            elif any(key in ep_keys for key in {'prod', 'production', 'develop', 'development'}):
                ref = 'prod' if 'prod' in ep_keys else 'production'
                if self.is_development_env: ref = 'develop' if 'develop' in ep_keys else 'development'
                endpoints = self.endpoints[ref]

            if endpoints:
                if self.in_k8s and endpoints.get('api', {}).get('cluster') and \
                    validate_website_with_socket(endpoints['api']['cluster']):
                    config.server_url = endpoints['api']['cluster']
                    config.host_port = endpoints['api']['cluster']
                elif endpoints.get('grpc', {}).get('public'):
                    config.host_port = endpoints['grpc']['public']
            # refs = list(self.endpoints.get('api', {}).keys())
            # ref = 'develop' if self.is_development_env else 'prod' 
            # from lazyops.libs.abcs.utils.http import validate_website_with_socket
            # if self.in_k8s and self.endpoints.get('api', {}).get(ref, {}).get('cluster') and \
            #     validate_website_with_socket(
            #     self.endpoints['api'][ref]['cluster'],
            # ):
            #     config.server_url = self.endpoints['api'][ref]['cluster']
            #     config.host_port = self.endpoints['grpc'][ref]['cluster']
            elif self.endpoints.get('grpc', {}).get(ref, {}).get('external'):
                config.host_port = self.endpoints['grpc'][ref]['external']
            
        return config


    def configure_session_endpoints(
        self, 
        config: 'ClientConfig',
        instance: str = 'default',
        api_endpoint: Optional[str] = None,
        grpc_endpoint: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Configure the session endpoints based on the environment variables
        """
        # If no endpoints are provided, then we use the endpoints from the config
        if not api_endpoint and os.getenv(f'HATCHET_API_ENDPOINT_{instance.upper()}'):
            api_endpoint = os.getenv(f'HATCHET_API_ENDPOINT_{instance.upper()}')
        
        if not grpc_endpoint and os.getenv(f'HATCHET_GRPC_ENDPOINT_{instance.upper()}'):
            grpc_endpoint = os.getenv(f'HATCHET_GRPC_ENDPOINT_{instance.upper()}')

        if not api_endpoint and not grpc_endpoint and (
            not self.endpoints and not \
            self.endpoints.get(instance)
        ):
            return self.configure_client_endpoints(config)
        if api_endpoint and grpc_endpoint:
            config.server_url = api_endpoint
            config.host_port = grpc_endpoint
        elif grpc_endpoint:
            config.host_port = grpc_endpoint
        elif self.endpoints.get(instance):
            endpoints = self.endpoints[instance]
            ref = 'develop' if self.is_development_env else 'prod' 
            from lazyops.libs.abcs.utils.http import validate_website_with_socket
            if self.in_k8s and endpoints.get('api', {}).get(ref, {}).get('cluster') and \
                validate_website_with_socket(
                endpoints['api'][ref]['cluster'],
            ):
                config.server_url = endpoints['api'][ref]['cluster']
                config.host_port = endpoints['grpc'][ref]['cluster']
            elif endpoints.get('grpc', {}).get(ref, {}).get('external'):
                config.host_port = endpoints['grpc'][ref]['external']
        elif self.endpoints:
            ep_keys = list(self.endpoints.keys())
            from lazyops.libs.abcs.utils.http import validate_website_with_socket
            endpoints: Dict[str, Dict[str, str]] = None
            if any(key in ep_keys for key in {'api', 'grpc'}):
                refs = list(self.endpoints.get('api', {}).keys())
                # This means there is not env config
                if 'cluster' in refs or 'public' in refs:
                    endpoints = self.endpoints
                else:
                    ref = 'prod' if 'prod' in refs else 'production'
                    if self.is_development_env: ref = 'develop' if 'develop' in refs else 'development'
                    endpoints = {
                        'api': self.endpoints['api'][ref],
                        'grpc': self.endpoints.get('grpc', {}).get(ref),
                    }
            elif any(key in ep_keys for key in {'prod', 'production', 'develop', 'development'}):
                ref = 'prod' if 'prod' in ep_keys else 'production'
                if self.is_development_env: ref = 'develop' if 'develop' in ep_keys else 'development'
                endpoints = self.endpoints[ref]

            if endpoints:
                # self.logger.info(f'Configuring Hatchet Endpoints: {endpoints}', colored = True)
                if self.in_k8s and endpoints.get('api', {}).get('cluster') and \
                    validate_website_with_socket(endpoints['api']['cluster']):
                    config.server_url = endpoints['api']['cluster']
                    config.host_port = endpoints['grpc']['cluster']
                elif endpoints.get('grpc', {}).get('public'):
                    config.host_port = endpoints['grpc']['public']
            elif self.endpoints.get('grpc', {}).get(ref, {}).get('external'):
                config.host_port = self.endpoints['grpc'][ref]['external']
        return config

    def configure_session_endpoints_v1(
        self, 
        config: 'ClientConfig',
        instance: str = 'default',
        api_endpoint: Optional[str] = None,
        grpc_endpoint: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Configure the session endpoints based on the environment variables
        """
        # If no endpoints are provided, then we use the endpoints from the config
        if not api_endpoint and os.getenv(f'HATCHET_API_ENDPOINT_{instance.upper()}'):
            api_endpoint = os.getenv(f'HATCHET_API_ENDPOINT_{instance.upper()}')
        
        if not grpc_endpoint and os.getenv(f'HATCHET_GRPC_ENDPOINT_{instance.upper()}'):
            grpc_endpoint = os.getenv(f'HATCHET_GRPC_ENDPOINT_{instance.upper()}')
        
        ref = 'develop' if self.is_development_env else 'production' 
        if not api_endpoint and not grpc_endpoint and (
            not self.endpoints or not \
            self.endpoints.get(instance)
        ):
            return self.configure_client_endpoints(config)
        if api_endpoint and grpc_endpoint:
            config.server_url = api_endpoint
            config.host_port = grpc_endpoint
        elif grpc_endpoint:
            config.host_port = grpc_endpoint
        elif self.endpoints.get(ref):
            endpoints = self.endpoints[ref]
            if endpoints.get('api', {}).get('cluster') and \
                validate_website_with_socket(endpoints['api']['cluster']):
                config.server_url = endpoints['api']['cluster']
                config.host_port = endpoints['grpc']['cluster']
            elif endpoints.get('grpc', {}).get('public'):
                config.host_port = endpoints['grpc']['public']
        elif self.endpoints.get(instance):
            endpoints = self.endpoints[instance]
            ref = 'develop' if self.is_development_env else 'production' 
            from lazyops.libs.abcs.utils.http import validate_website_with_socket
            if self.in_k8s and endpoints.get('api', {}).get(ref, {}).get('cluster') and \
                validate_website_with_socket(
                endpoints['api'][ref]['cluster'],
            ):
                config.server_url = endpoints['api'][ref]['cluster']
                config.host_port = endpoints['grpc'][ref]['cluster']
            elif endpoints.get('grpc', {}).get(ref, {}).get('external'):
                config.host_port = endpoints['grpc'][ref]['external']
        elif self.endpoints.get('api', {}).get('cluster') and \
            validate_website_with_socket(self.endpoints['api']['cluster']):
            config.server_url = self.endpoints['api']['cluster']
            config.host_port = self.endpoints['grpc']['cluster']
        elif self.endpoints.get('grpc', {}).get('public'):
            config.host_port = self.endpoints['grpc']['public']
        return config


    @property
    def logger(self) -> 'Logger':
        """
        Gets the logger
        """
        return logger
    
    @property
    def null_logger(self) -> 'Logger':
        """
        Gets the null logger
        """
        return null_logger
    
    @property
    def autologger(self) -> 'Logger':
        """
        Automatic Logger that is enabled in devel mode
        """
        return logger if (self.is_development_env or self.debug_enabled) else null_logger
    
    @property
    def version(self) -> str:
        """
        Returns the hatchet sdk version
        """
        if 'version' not in self._extra:
            from importlib.metadata import version
            self._extra['version'] = version('hatchet-sdk')
        return self._extra['version']
    
    @version.setter
    def version(self, value: str):
        """
        Sets the hatchet sdk version
        """
        self._extra['version'] = value

    def configure(self, **kwargs):
        """
        Configures the Hatchet Settings
        """
        for k,v in kwargs.items():
            if k == 'app_env':
                if isinstance(v, str): v = AppEnv(v)
            if hasattr(self, k): setattr(self, k, v)
        return self
        

# settings: HatchetSettings = ProxyObject(
#     obj_getter = 'lazyops.libs.hatchet.utils.get_hatchet_settings'
# )


from __future__ import annotations

"""
Formatting Helpers
"""

import re
import html
import datetime
import contextlib
from typing import List, Dict, Any, Union, Optional, Tuple

def build_dict_from_str(
    data: str,
    **kwargs
) -> Union[List[Any], Dict[str, Any]]:
    """
    Helper to build a dictionary from a string
    """
    import json
    if (data.startswith('[') and data.endswith(']')) or (data.startswith('{') and data.endswith('}')):
        return json.loads(data)
    return build_dict_from_list(data.split(','), **kwargs)


def build_dict_from_list(
    data: List[str],
    seperator: str = '=',
) -> Dict[str, Any]:
    """
    Builds a dictionary from a list of strings
    """
    import json
    return json.loads(str(dict([item.split(seperator) for item in data])))


def build_dict_from_query(
    query: str,
    **kwargs,
) -> Dict[str, Union[int, float, str, datetime.datetime, Any]]:
    """
    Builds a dictionary from a query
    """
    # Try to base decode it
    if not query.startswith('{') and not query.startswith('['):
        import base64
        query = base64.b64decode(query).decode('utf-8')
    data = build_dict_from_str(query, **kwargs)
    for k,v in data.items():
        if 'date' in k:
            with contextlib.suppress(Exception):
                from lzo.utils.helpers.dates import parse_datetime
                v = parse_datetime(v)
        data[k] = v
    return data


## Cleaning

html_re_pattern = re.compile('<.*?>|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});')
alphanumeric_pattern = re.compile(r'[^a-zA-Z0-9]')
dba_pattern = re.compile(r'(.+?)(?:\s+d[/]b[/]a|[/]d[/]b[/]a|[/]dba|[/]DBA)(?:\s+(.+))?$')
url_pattern = re.compile(r'(https?://\S+)')
name_pattern_regex = re.compile(r'(?:(?<=^)|(?<=[^A-Za-z.,]))[A-Za-z.,]+(?: [A-Za-z.,]+)*(?:(?=[^A-Za-z.,])|(?=$))')
underscore_pattern = re.compile("___+")

def clean_html(text: str) -> str:
    """
    Clean HTML
    """
    return html_re_pattern.sub('', html.unescape(text))


def cleanup_dots(text: str) -> str:
    """
    Cleans up the dots found in table of contents

    INTRODUCTION............................................ > INTRODUCTION.
    """
    return re.sub(r'\.+', '.', text).strip()

def extract_urls(text: str) -> List[str]:
    """
    Extracts urls from a string
    """
    return url_pattern.findall(text)

def cleanup_whitespace(text: str) -> str:
    """
    Cleans up the whitespace
    """
    return re.sub(r'\s+', ' ', text).strip()

def format_for_id(text: str, sep: str = '-') -> str:
    """
    Formats the text for an ID
    """
    return re.sub(r'\W+', sep, text).strip(sep).lower()


def extract_email(text: str) -> Optional[str]:
    """
    Extracts the email
    """
    match = re.search(r'[\w\.-]+@[\w\.-]+', text)
    return match[0] if match else None

def extract_phone(text: str) -> Optional[str]:
    """
    Extracts the phone
    """
    # 205460001
    # 3014430995
    match = re.search(r'[\d\(\)\-]+', text)
    if not match: return None
    pn = match[0]
    if len(pn.replace('(', '').replace(')', '').replace('-', '').replace(' ', '')) <= 9: return None
    return pn

def extract_name(text: str) -> Optional[str]:
    """
    Extracts the name
    """
    text = text.replace('Contact', '').replace('at', '').strip()
    match = name_pattern_regex.search(text)
    return match[0] if match else None
    
def title_case_text(text: str) -> str:
    """
    Title cases the text
    """
    return ' '.join([word.capitalize() for word in text.split(' ')])

def unescape_html(text: str, replace_breaks: Optional[bool] = True) -> str:
    """
    Unescape HTML
    -> Kimberlie J. Laing, Grant Analyst, 301-903-3026&lt;br/&gt;kim.laing@science.doe.gov&lt;br/&gt;Catherine M. Ronning, Program Manager, 301-903-9549&lt;br/&gt;Catherine.Ronning@science.doe.gov&lt;br/&gt;'
    -> Kimberlie J. Laing, Grant Analyst, 301-903-3026<br/>kim.laing@science.doe.gov<br/>Catherine M. Ronning, Program Manager, 301-903-9549<br/>Catherine.Ronning@science.doe.gov<br/>
    """
    text = text.replace('�', ' ')
    text = html.unescape(text)
    if replace_breaks: text = text.replace('<br/>', '\n')
    return text.strip()


def normalize_sql_text(text: Optional[str]) -> Optional[str]:
    """
    Normalizes the text to be indexed/queried
    """
    if text is None or not text.strip(): return None
    return alphanumeric_pattern.sub(' ', text).replace('  ', ' ').replace('  ', ' ').strip()


def strip_or_none(text: str) -> Optional[str]:
    """
    Strips the text or returns None
    """
    return text.strip() or None

def extract_dba_and_company_name(company_name: str) -> Tuple[str, str]:
    # sourcery skip: extract-method
    """
    Extracts the DBA from the company name

    Examples:
        - SafeFlights Inc /D/B/A 14bis Supply Tracking -> [SafeFlights Inc, 14bis Supply Tracking]
        - compact power inc of america d/b/a Lightening Energy -> [compact power inc of america, Lightening Energy]
        - TaskUnite Inc (dba AMPAworks) -> [TaskUnite Inc, AMPAworks]
        - Systems And Processes Engineering Corporation (SPEC) -> [Systems And Processes Engineering Corporation, SPEC]
        - FitChimp Inc (DBA FitRankings) -> [FitChimp Inc, FitRankings]
        - (ES3) Engineering AND Software System Solution Inc -> [ES3, Engineering AND Software System Solution Inc]
        - Opto-Knowledge Systems Inc (OKSI) -> [Opto-Knowledge Systems Inc, OKSI]
        - Protonex LLC dba PNI Sensor -> [Protonex LLC, PNI Sensor]
        - AIR REFUELING TANKER TRANSPORT (ART2) LLC -> [AIR REFUELING TANKER TRANSPORT, ART2]
    """

    # Use re.match to find the pattern in the company name
    match = dba_pattern.match(company_name, re.IGNORECASE)
    if match:
        # Extract the Dba and company name parts
        dba = match.group(1).strip()
        company = match.group(2).strip() if match.group(2) else None
        if company and '(' in company: company = company[:company.find('(')-1].strip()
        return [strip_or_none(dba), strip_or_none(company)]

    i_start, i_end = company_name.find('('), company_name.find(')')
    if i_start != -1 and i_end != -1:
        dba = company_name[i_start+1:i_end].replace('dba', '').replace('DBA', '').strip()
        company = company_name.replace(dba, '').replace('(', '').replace(')', '').replace('dba', '').replace('DBA', '').replace('  ', ' ').strip()
        return [strip_or_none(dba), strip_or_none(company)]

    if 'dba' in company_name:
        parts = company_name.split('dba')
        return [strip_or_none(parts[1]), strip_or_none(parts[0])]
    
    return [company_name.strip(), None]


def count_non_sequential_integers(input_string: str) -> int:
    """
    Determine the number of non-sequential integers in a given string.
    
    Args:
        input_string (str): The input string to be analyzed.
        
    Returns:
        int: The number of non-sequential integers in the input string.
    """
    # Find all integers in the input string
    integers = [int(x) for x in re.findall(r'\d+', input_string)]
    return sum(integers[i] != integers[i - 1] + 1 for i in range(1, len(integers)))


def is_spam_domain(domain: str) -> bool:
    """
    Identify if a given domain is likely a spam domain.
    """
    if len(domain) > 50: return True
    if domain.count('i') > 4 and domain.count('a') > 4: return True
    if (
        domain.count('x') > 2 or 
        domain.count('z') > 2 or 
        domain.count('j') > 3 or 
        domain.count('w') > 3 or 
        domain.count('y') > 3 or 
        domain.count('f') > 3
    ): return True

    # Test Combinations
    if domain.count('x') > 1 and (domain.count('j') > 1 or domain.count('y') > 1): return True
    if domain.count('j') > 1 and domain.count('d') > 1 and domain.count('a') > 1: return True
    
    # Test for non-sequential integers
    n_non_seq = count_non_sequential_integers(domain)
    # if n_non_seq > 1:  return True
    return n_non_seq > 1
    

def reformat_spacing(text: str) -> str:
    """
    Reformat the spacing
    """
    text = text.replace('\n[', '\n\n**').replace(']:', ':**').replace('[', '**')
    text = text.replace('\n  - ', '\n\n  - ').replace('\n- ', '\n\n- ')
    return text.strip()

def parse_duration(duration: str) -> float:
    """
    Parses the duration

    5 minutes, 49.75 seconds -> 349.75
    """
    total = 0.0
    parts = duration.split(', ')
    for part in parts:
        if ' seconds' in part: total += float(part.split(' ')[0])
        elif ' minutes' in part: total += float(part.split(' ')[0]) * 60
        elif ' hours' in part: total += float(part.split(' ')[0]) * 3600
        elif ' days' in part: total += float(part.split(' ')[0]) * 86400
    return total


def combine_parts(*parts: Optional[str], sep: Optional[str] = '.') -> str:
    """
    Combines the parts into a single string
    """
    return sep.join(p for p in parts if p)


def clean_text(text: str) -> str:
    """
    Cleans the text
    """
    if '�' in text: text = text.replace('�', '')
    
    # Find and remove double underscores
    if underscore_pattern.search(text):
        text = underscore_pattern.sub(' ', text)
    
    # Cleanup HTML
    if '<' in text and '>' in text: 
        with contextlib.suppress(Exception):
            from markdownify import markdownify
            text = markdownify(text)
    
    return cleanup_whitespace(text)


"""
Miscellaneous Types
"""


import typing as t
from lzl.types import PYDANTIC_VERSION
from pydantic.types import ByteSize
try:
    import dill as pickle
except ImportError:
    import pickle

class ObjectSize(ByteSize):
    """
    Returns the size of an object
    """

    @classmethod
    def _dump_(cls, obj: t.Any) -> float:
        """
        Returns the size of the object
        """
        return len(obj) if isinstance(obj, (bytes, bytearray)) else len(pickle.dumps(obj))

    if PYDANTIC_VERSION == 2:
        from pydantic_core import core_schema

        @classmethod
        def _validate(cls, input_value: t.Any, /, _: core_schema.ValidationInfo) -> 'ObjectSize':
            """
            Validator for Pydantic v2
            """
            if not isinstance(input_value, (int, float)):
                input_value = cls._dump_(input_value)
            return super()._validate(input_value, _)

    else:
        @classmethod
        def __get_validators__(cls):
            yield cls.validate
        
        @classmethod
        def validate(cls, input_value: t.Any) -> 'ObjectSize':
            """
            Validator for Pydantic v1
            """
            if not isinstance(input_value, (int, float)):
                input_value = cls._dump_(input_value)
            return super()._validate(input_value)

    def __new__(cls, obj: t.Any):
        """
        Returns the size of the object
        """
        if not isinstance(obj, (int, float)):
            obj = cls._dump_(obj)
        return super().__new__(cls, obj)


from __future__ import annotations

"""Context helpers for tracking system metrics during worker execution."""

import abc
import contextlib
import gc
import typing as t

from lzl.types import eproperty

if t.TYPE_CHECKING:
    from lzo.utils.system.gpu import GPUData
    from lzo.utils.system.resources import ResourceData


class WorkerContext(abc.ABC):
    """Capture CPU/GPU statistics around long-running worker operations."""

    def __init__(self, **kwargs: t.Any) -> None:
        """Initialise timer/logging helpers used during metric collection."""

        from lzl.logging import logger
        from lzo.utils import Timer
        from lzo.utils.system import aget_gpu_data, get_gpu_data, get_resource_data
        from pydantic.types import ByteSize

        self._extra: dict[str, t.Any] = {}
        self._kwargs = kwargs

        self.timer = Timer
        self.logger = logger
        self._bs = ByteSize

        self.get_gpu_data = get_gpu_data
        self.aget_gpu_data = aget_gpu_data
        self.get_resource_data = get_resource_data

        self.t = self.timer()
        self.idx: int = 0
        self.num_batches: int = 0
        self.last_batch_size: int = 0
        self.total_duration: float = 0.0
        self.last_duration: float = 0.0

        self.last_resource_data: t.Optional["ResourceData"] = None
        self.last_gpu_data: t.Optional["GPUData"] = None

    # ------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------

    @eproperty
    def torch_device_name(self) -> str:
        """Return the active PyTorch device name."""

        from lzo.utils.system import get_torch_device_name

        return get_torch_device_name()

    @eproperty
    def torch_device(self):
        """Return the active PyTorch device instance."""

        from lzo.utils.system import get_torch_device

        return get_torch_device()

    @eproperty
    def has_gpu(self) -> bool:
        """Return ``True`` when the runtime is backed by CUDA."""

        return self.torch_device_name.startswith("cuda")

    @eproperty
    def model_name(self) -> t.Optional[str]:
        """Optional model identifier surfaced in log prefixes."""

        return self._extra.get("model_name")

    @eproperty
    def worker_name(self) -> t.Optional[str]:
        """Optional worker identifier used in log prefixes."""

        return self._extra.get("worker_name")

    # ------------------------------------------------------------------
    # GPU helpers
    # ------------------------------------------------------------------

    def build_gpu_data_string(
        self,
        current_usage: "GPUData",
        compare: bool | None = None,
        previous_usage: "GPUData" | None = None,
        colored: bool = False,
    ) -> t.Optional[str]:
        """Format GPU usage statistics with optional comparison."""

        curr_mem_used = current_usage["memory_used"]
        curr_mem_percent = current_usage["utilization_memory"]
        curr_mem_total = current_usage["memory_total"]
        gpu_name = current_usage["name"]

        previous_usage = previous_usage or self.last_gpu_data
        self.last_gpu_data = current_usage

        if compare and previous_usage:
            comparison = {
                "memory_used": self._bs(curr_mem_used - previous_usage["memory_used"]),
                "utilization_memory": curr_mem_percent - previous_usage["utilization_memory"],
            }
            if not colored:
                return (
                    f"{gpu_name}: {previous_usage['memory_used'].human_readable()} -> "
                    f"{curr_mem_used.human_readable()} / {curr_mem_total.human_readable()} + "
                    f"{comparison['memory_used'].human_readable()} "
                    f"({comparison['utilization_memory']} -> {curr_mem_percent}%)"
                )
            return (
                f"{gpu_name}: {previous_usage['memory_used'].human_readable()} -> |y|{curr_mem_used.human_readable()}|e| / "
                f"|g|{curr_mem_total.human_readable()}|e| + |r|{comparison['memory_used'].human_readable()}|e| "
                f"({comparison['utilization_memory']} -> {curr_mem_percent}%)"
            )

        if not colored:
            return (
                f"{gpu_name}: {curr_mem_used.human_readable()} / {curr_mem_total.human_readable()} "
                f"({curr_mem_percent}%)"
            )
        return (
            f"{gpu_name}: |y|{curr_mem_used.human_readable()}|e| / |g|{curr_mem_total.human_readable()}|e| "
            f"({curr_mem_percent}%)"
        )

    def get_gpu_memory(
        self,
        compare: bool | None = None,
        previous_usage: "GPUData" | None = None,
        colored: bool = False,
    ) -> t.Optional[str]:
        """Return formatted GPU usage, optionally comparing against a prior sample."""

        current_usage = self.get_gpu_data()
        if not current_usage:
            return None
        return self.build_gpu_data_string(
            current_usage,
            compare=compare,
            previous_usage=previous_usage,
            colored=colored,
        )

    async def aget_gpu_memory(
        self,
        compare: bool | None = None,
        previous_usage: "GPUData" | None = None,
        colored: bool = False,
    ) -> t.Optional[str]:
        """Asynchronous variant of :meth:`get_gpu_memory`."""

        current_usage = await self.aget_gpu_data()
        if not current_usage:
            return None
        return self.build_gpu_data_string(
            current_usage,
            compare=compare,
            previous_usage=previous_usage,
            colored=colored,
        )

    # ------------------------------------------------------------------
    # Resource helpers
    # ------------------------------------------------------------------

    def build_resource_data_string(
        self,
        current_usage: "ResourceData",
        compare: bool | None = None,
        previous_usage: "ResourceData" | None = None,
        colored: bool = False,
    ) -> t.Optional[str]:
        """Format CPU/RAM usage into a human-readable string."""

        curr_mem_used = current_usage["memory_used"]
        curr_mem_percent = current_usage["utilization_memory"]
        curr_mem_total = current_usage["memory_total"]
        curr_cpu_percent = current_usage["utilization_cpu"]
        num_cpu = current_usage["cpu_count"]

        previous_usage = previous_usage or self.last_resource_data
        self.last_resource_data = current_usage

        if compare and previous_usage:
            comparison = {
                "memory_used": self._bs(curr_mem_used - previous_usage["memory_used"]),
            }
            base = (
                f"{num_cpu} vCPU: {curr_cpu_percent}% | RAM: {previous_usage['memory_used'].human_readable()} -> "
                f"{curr_mem_used.human_readable()} / {curr_mem_total.human_readable()} + "
                f"{comparison['memory_used'].human_readable()} ({curr_mem_percent}%)"
            )
            if not colored:
                return base
            return (
                f"{num_cpu} vCPU: |g|{curr_cpu_percent}%|e| | RAM: {previous_usage['memory_used'].human_readable()} -> "
                f"|y|{curr_mem_used.human_readable()}|e| / |g|{curr_mem_total.human_readable()}|e| + "
                f"|r|{comparison['memory_used'].human_readable()}|e| ({curr_mem_percent}%)"
            )

        if not colored:
            return (
                f"{num_cpu} vCPU: {curr_cpu_percent}% | RAM: {curr_mem_used.human_readable()} / "
                f"{curr_mem_total.human_readable()} ({curr_mem_percent}%)"
            )
        return (
            f"{num_cpu} vCPU: |g|{curr_cpu_percent}%|e| RAM: |y|{curr_mem_used.human_readable()}|e| / "
            f"|g|{curr_mem_total.human_readable()}|e| ({curr_mem_percent}%)"
        )

    def get_resource_info(
        self,
        compare: bool | None = None,
        previous_usage: "ResourceData" | None = None,
        colored: bool = False,
    ) -> t.Optional[str]:
        """Return formatted CPU/memory usage snapshot."""

        current_usage = self.get_resource_data()
        if not current_usage:
            return None
        return self.build_resource_data_string(
            current_usage,
            compare=compare,
            previous_usage=previous_usage,
            colored=colored,
        )
    # ------------------------------------------------------------------
    # Context managers
    # ------------------------------------------------------------------

    @contextlib.contextmanager
    def inference_mode(
        self,
        batch_size: int | None = 1,
        obj_name: str | None = None,
        enable_gc: bool | None = None,
        enable_summary: bool | None = None,
        hook: t.Optional[t.Callable[[dict[str, t.Any]], None]] = None,
        **kwargs: t.Any,
    ):
        """Wrap a block of inference work with resource logging."""

        ts = self.timer(format_ms=True, format_short=1)
        start_text = "Starting Inference"
        if obj_name:
            start_text += f" for |g|{obj_name}|e|"
        start_text += f" ({batch_size})"
        self.logger.info(start_text, prefix=self.model_name, colored=True, hook=hook)
        start_resource_data = self.get_resource_data()
        start_gpu_data = self.get_gpu_data() if self.has_gpu else None
        try:
            yield
        except Exception as exc:  # pragma: no cover - logging path
            self.logger.trace(f"[{self.model_name}] Error in Inference Mode: ", exc, hook=hook)
            raise
        finally:
            total_s = ts.total
            self.total_duration += total_s
            self.last_duration = total_s
            self.num_batches += batch_size or 0
            self.idx += 1

            end_text = "Inference Completed"
            if obj_name:
                end_text += f" for |g|{obj_name}|e|"
            end_text += f" ({batch_size}) in {ts.total_s}"
            if enable_gc:
                gc.collect()

            self.logger.info(end_text, colored=True, prefix=self.model_name, hook=hook)
            if enable_summary:
                self.logger.info(
                    f"Total Requests: |g|{self.idx}|e|. Total Batches Handled: |g|{self.num_batches}|e|. "
                    f"Handled Last Batch Size of |g|{self.last_batch_size}|e|",
                    colored=True,
                    prefix=self.model_name,
                    hook=hook,
                )
                self.logger.info(
                    f"Total Inference Duration: |g|{self.timer.pformat_duration(self.total_duration)}|e|. "
                    f"Total Time Alive: |g|{self.t.total_s}|e|",
                    colored=True,
                    prefix=self.model_name,
                    hook=hook,
                )
            self.logger.info(
                self.get_resource_info(compare=True, previous_usage=start_resource_data, colored=True),
                colored=True,
                prefix=self.model_name,
                hook=hook,
            )
            if self.has_gpu and start_gpu_data is not None:
                self.logger.info(
                    self.get_gpu_memory(compare=True, previous_usage=start_gpu_data, colored=True),
                    colored=True,
                    prefix=self.model_name,
                    hook=hook,
                )

    @contextlib.asynccontextmanager
    async def ainference_mode(
        self,
        batch_size: int | None = 1,
        obj_name: str | None = None,
        enable_gc: bool | None = None,
        enable_summary: bool | None = None,
        hook: t.Optional[t.Callable[[dict[str, t.Any]], None]] = None,
        **kwargs: t.Any,
    ):
        """Async variant of :meth:`inference_mode`."""

        ts = self.timer(format_ms=True, format_short=1)
        start_text = "Starting Inference"
        if obj_name:
            start_text += f" for |g|{obj_name}|e|"
        start_text += f" ({batch_size})"
        self.logger.info(start_text, prefix=self.model_name, colored=True, hook=hook)
        start_resource_data = self.get_resource_data()
        start_gpu_data = await self.aget_gpu_data() if self.has_gpu else None
        try:
            yield
        except Exception as exc:  # pragma: no cover - logging path
            self.logger.trace(f"[{self.model_name}] Error in Inference Mode: ", exc, hook=hook)
            raise
        finally:
            total_s = ts.total
            self.total_duration += total_s
            self.last_duration = total_s
            self.num_batches += batch_size or 0
            self.idx += 1

            end_text = "Inference Completed"
            if obj_name:
                end_text += f" for |g|{obj_name}|e|"
            end_text += f" ({batch_size}) in {ts.total_s}"
            if enable_gc:
                gc.collect()

            self.logger.info(end_text, colored=True, prefix=self.model_name, hook=hook)
            if enable_summary:
                self.logger.info(
                    f"Total Requests: |g|{self.idx}|e|. Total Batches Handled: |g|{self.num_batches}|e|. "
                    f"Handled Last Batch Size of |g|{self.last_batch_size}|e|",
                    colored=True,
                    prefix=self.model_name,
                    hook=hook,
                )
                self.logger.info(
                    f"Total Inference Duration: |g|{self.timer.pformat_duration(self.total_duration)}|e|. "
                    f"Total Time Alive: |g|{self.t.total_s}|e|",
                    colored=True,
                    prefix=self.model_name,
                    hook=hook,
                )
            self.logger.info(
                self.get_resource_info(compare=True, previous_usage=start_resource_data, colored=True),
                colored=True,
                prefix=self.model_name,
                hook=hook,
            )
            if self.has_gpu and start_gpu_data is not None:
                self.logger.info(
                    await self.aget_gpu_memory(compare=True, previous_usage=start_gpu_data, colored=True),
                    colored=True,
                    prefix=self.model_name,
                    hook=hook,
                )

    @contextlib.contextmanager
    def capture(
        self,
        message: str | None = None,
        prefix: str | None = None,
        hook: t.Optional[t.Callable[[dict[str, t.Any]], None]] = None,
        **kwargs: t.Any,
    ):
        """Capture resources across an arbitrary block of work."""

        ts = self.timer(format_ms=True, format_short=1)
        base_name = self.model_name or self.worker_name
        prefix = f"{prefix} {base_name}" if prefix else base_name
        start_resource_data = self.get_resource_data()
        start_gpu_data = self.get_gpu_data() if self.has_gpu else None
        try:
            yield
        except Exception as exc:  # pragma: no cover - logging path
            self.logger.trace(f"[{prefix}] Error in Capture: ", exc, hook=hook)
            raise
        finally:
            message = (message or "Capture Complete") + f" in {ts.total_s}"
            self.logger.info(message, colored=True, prefix=prefix, hook=hook)
            self.logger.info(
                self.get_resource_info(compare=True, previous_usage=start_resource_data, colored=True),
                colored=True,
                prefix=prefix,
                hook=hook,
            )
            if self.has_gpu and start_gpu_data is not None:
                self.logger.info(
                    self.get_gpu_memory(compare=True, previous_usage=start_gpu_data, colored=True),
                    colored=True,
                    prefix=prefix,
                    hook=hook,
                )

    @contextlib.contextmanager
    def start_task(
        self,
        batch_size: int | None = 1,
        obj_name: str | None = None,
        task_name: str | None = None,
        enable_gc: bool | None = None,
        enable_summary: bool | None = None,
        hook: t.Optional[t.Callable[[dict[str, t.Any]], None]] = None,
        **kwargs: t.Any,
    ):
        """Track a generic worker task, logging CPU/GPU usage."""

        ts = self.timer(format_ms=True, format_short=1)
        base_name = self.model_name or self.worker_name
        start_text = "Starting Task"
        if task_name:
            start_text += f": |g|{task_name}|e|"
        if obj_name:
            start_text += f" for |g|{obj_name}|e|"
        start_text += f" ({batch_size})"
        self.logger.info(start_text, prefix=base_name, colored=True, hook=hook)
        start_gpu_data = self.get_gpu_data() if self.has_gpu else None
        try:
            yield
        except Exception as exc:  # pragma: no cover - logging path
            self.logger.trace(f"[{base_name}] Error in Task: ", exc, hook=hook)
            raise
        finally:
            total_s = ts.total
            self.total_duration += total_s
            self.last_duration = total_s
            self.num_batches += batch_size or 0
            self.idx += 1

            end_text = "Task Completed"
            if task_name:
                end_text += f": |g|{task_name}|e|"
            if obj_name:
                end_text += f" for |g|{obj_name}|e|"
            end_text += f" ({batch_size}) in {ts.total_s}"
            if enable_gc:
                gc.collect()

            self.logger.info(end_text, colored=True, prefix=base_name, hook=hook)
            if enable_summary:
                self.logger.info(
                    f"Total Tasks: |g|{self.idx}|e|. Total Batches Handled: |g|{self.num_batches}|e|. "
                    f"Handled Last Batch Size of |g|{self.last_batch_size}|e|",
                    colored=True,
                    prefix=base_name,
                    hook=hook,
                )
                self.logger.info(
                    f"Total Task Duration: |g|{self.timer.pformat_duration(self.total_duration)}|e|. "
                    f"Total Time Alive: |g|{self.t.total_s}|e|",
                    colored=True,
                    prefix=base_name,
                    hook=hook,
                )
            self.logger.info(
                self.get_resource_info(compare=False, colored=True),
                colored=True,
                prefix=base_name,
                hook=hook,
            )
            if self.has_gpu and start_gpu_data is not None:
                self.logger.info(
                    self.get_gpu_memory(compare=True, previous_usage=start_gpu_data, colored=True),
                    colored=True,
                    prefix=base_name,
                    hook=hook,
                )

    def __enter__(self) -> "WorkerContext":
        return self

    def __exit__(self, exc_type, exc, tb) -> None:
        return None



from __future__ import annotations

"""Utility façade aggregating commonly used helpers within ``lzo``."""

from .logs import (
    NullLogger,
    Logger,
    
    logger,
    null_logger,

    change_logger_level,
    add_api_log_filters,
    get_logger,
    
)
from .helpers.base import (
    extract_function_kwargs,
    is_coro_func,
    create_secret,
    create_unique_id,

    update_dict,
    merge_recursive_dict,
    flatten_dict_value,

    fail_after,
)

from .helpers.timing import (
    Timer,
    timer,
    DynamicInterval,
)


from .helpers.dates import (
    convert_date,
    parse_datetime,
    parse_datetime_from_timestamp,
    get_current_datetime,
    is_expired_datetime,
    create_timestamp,
)

from .helpers.caching import (
    timed_cache,
    is_in_async_loop,
)

from .helpers.formatting import (
    build_dict_from_str,
    build_dict_from_list,
    build_dict_from_query,
    combine_parts,
)

from .helpers.envvars import (
    load_env_vars,
    parse_envvars_from_text,
    parse_from_envvar,
)

__all__ = [
    'NullLogger',
    'Logger',
    'add_api_log_filters',
    'change_logger_level',
    'create_secret',
    'create_unique_id',
    'extract_function_kwargs',
    'fail_after',
    'get_logger',
    'logger',
    'null_logger',
    'timer',
]



from __future__ import annotations

"""
SQLModel Extended Types
"""


# from sqlmodel import SQLModel as BaseSQLModel
import functools
from sqlmodel._compat import get_annotations
from pydantic._internal._repr import Representation
from sqlmodel.main import SQLModelMetaclass, Field, FieldInfo, SQLModelConfig, SQLModel as BaseSQLModel
from sqlmodel.main import default_registry
from lzl.types import eproperty, PrivateAttr
from typing import Any, ClassVar, Coroutine, List, Dict, Tuple, TypeVar, Type, Awaitable, Callable, Set, Optional, Union, ClassVar, TYPE_CHECKING
from sqlalchemy.util.concurrency import greenlet_spawn


def schema_extra(schema: dict, _):
    props = {
        k: v
        for k, v in schema.get('properties', {}).items()
        if not v.get("hidden", False) and not k.startswith('_')
    }
    schema["properties"] = props

_T = TypeVar("_T")
def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]: # type: ignore
    return lambda a: a


class SQLModel(BaseSQLModel):
    """
    SQLModel with additional features
    """
    if not TYPE_CHECKING:
        _extra: Dict[str, Any] = PrivateAttr(default_factory = dict)
        model_config = SQLModelConfig(from_attributes = True, arbitrary_types_allowed = True, json_schema_extra = schema_extra)


_sql_model_init_hooks: Set[Callable[..., Any]] = set()

def add_sql_model_init_hook(func: Callable[..., Any]) -> None:
    """
    Adds an SQLModel Init Hook
    """
    _sql_model_init_hooks.add(func)

def run_sql_model_init_hooks(model: Type['SQLModel']) -> None:
    """
    Runs the SQLModel Init Hooks
    """
    for hook in _sql_model_init_hooks:
        hook(model)

# Adding AsyncSQLModel to the list of SQLModels
# https://github.com/fastapi/sqlmodel/pull/872/files

class AwaitableFieldInfo(Representation):
    def __init__(self, *, field: str):
        self.field = field


def AwaitableField(*, field: str) -> Any:
    return AwaitableFieldInfo(field=field)

@__dataclass_transform__(kw_only_default=True, field_descriptors=(Field, FieldInfo))
class AsyncSQLModelMetaclass(SQLModelMetaclass):
    __async_sqlmodel_awaitable_fields__: Dict[str, AwaitableFieldInfo]

    def __new__(
        cls,
        name: str,
        bases: Tuple[Type[Any], ...],
        class_dict: Dict[str, Any],
        **kwargs: Any,
    ) -> Any:
        awaitable_fields: Dict[str, AwaitableFieldInfo] = {}
        dict_for_sqlmodel = {}
        original_annotations = get_annotations(class_dict)
        sqlmodel_annotations = {}
        awaitable_fields_annotations = {}
        for k, v in class_dict.items():
            if isinstance(v, AwaitableFieldInfo):
                awaitable_fields[k] = v
            else:
                dict_for_sqlmodel[k] = v
        for k, v in original_annotations.items():
            if k in awaitable_fields:
                awaitable_fields_annotations[k] = v
            else:
                sqlmodel_annotations[k] = v

        dict_used = {
            **dict_for_sqlmodel,
            "__async_sqlmodel_awaitable_fields__": awaitable_fields,
            "__annotations__": sqlmodel_annotations,
        }
        return super().__new__(cls, name, bases, dict_used, **kwargs)

    def __init__(
        cls, classname: str, bases: Tuple[type, ...], dict_: Dict[str, Any], **kw: Any
    ) -> None:
        for field_name, field_info in cls.__async_sqlmodel_awaitable_fields__.items():

            def get_awaitable_field(
                self: "AsyncSQLModel", field: str = field_info.field
            ) -> Coroutine[Any, Any, Any]:
                return greenlet_spawn(getattr, self, field)

            setattr(cls, field_name, property(get_awaitable_field))
        
        SQLModelMetaclass.__init__(cls, classname, bases, dict_, **kw)
        run_sql_model_init_hooks(cls)

class AsyncSQLModel(SQLModel, metaclass=AsyncSQLModelMetaclass):
    
    __async_sqlmodel_awaitable_fields__: ClassVar[Dict[str, AwaitableFieldInfo]]


    async def __await_all_fields__(self):
        """
        Await all fields
        """
        field = None
        try:
            for field in self.__async_sqlmodel_awaitable_fields__:
                await getattr(self, field)
        except Exception as e:
            from lzl.logging import logger
            logger.trace(f'Error Awaiting Field: {field}', e)
            raise e

    def __await__(self):
        """
        Await the model fields
        """
        return self.__await_all_fields__().__await__()
    
_SM = TypeVar("_SM", bound=AsyncSQLModel)
_RM = TypeVar("_RM", bound=AsyncSQLModel)

def sqlm_method_wrapper(
    source: _SM,
    source_kws: Optional[List[str]] = None,
    inject_kwargs: Dict[str, Union[Callable, Any]] = None,
    exclude_none: Optional[bool] = True,
    method: Optional[str] = 'create',
    target_validator: Optional[str] = None,
) -> Callable[[_RM], _SM]:
    """
    Wrapper for SQLModel Methods
    """
    source_kws = source_kws or []
    def target_wrapper(target: _RM) -> Callable[[_RM], _SM]:
        """
        Target Wrapper
        """
        def inner_method(**kwargs: Any) -> _SM:
            """
            Inner Method
            """
            target_kws = {k:v for k,v in kwargs.items() if k not in source_kws}
            if exclude_none:
                target_kws = {k:v for k,v in target_kws.items() if v is not None}
            new_target = target(**target_kws)
            update_kwargs = {k:v for k,v in kwargs.items() if k in source_kws}
            if exclude_none:
                update_kwargs = {k:v for k,v in update_kwargs.items() if v is not None}
            if inject_kwargs:
                for k, v in inject_kwargs.items():
                    if callable(v): v = v()
                    update_kwargs[k] = v
            
            return source.model_validate(new_target, update = update_kwargs)
        
        setattr(target, method, inner_method)
        return target
        
    return target_wrapper




from __future__ import annotations

"""
Models for Argo Events
"""
import fnmatch
import typing as t

from pydantic import Field, AliasChoices
from lzl.types import BaseModel, eproperty
from pydantic import PrivateAttr, AliasChoices, ConfigDict, AliasGenerator, RootModel
from pydantic.alias_generators import to_camel

if t.TYPE_CHECKING:
    from lzl.io import File


def validation_gen(field: str) -> AliasChoices:
    """
    Generate validation aliases for a field.
    """
    return AliasChoices(to_camel(field), field)


default_alias_gen = AliasGenerator(
    validation_alias=validation_gen,
    serialization_alias=to_camel,
)


class BaseEventModel(BaseModel):
    """
    Base Argo Event Model
    """
    # The other stuff we don't really care about
    _extra: t.Dict[str, t.Any] = PrivateAttr(default_factory=dict)
    model_config = ConfigDict(
        extra = "allow",
        arbitrary_types_allowed = True,
        alias_generator = default_alias_gen,
        serialize_by_alias = True,
    )


class MinioBucket(BaseEventModel):
    """
    MinIO Bucket Data
    """
    name: t.Optional[str] = Field(default=None)
    owner_identity: t.Optional[dict] = Field(default=None)
    arn: t.Optional[str] = Field(default=None)


class MinioObject(BaseEventModel):
    """
    MinIO Object Data
    """
    key: t.Optional[str] = Field(default=None)
    size: t.Optional[int] = Field(default=None)
    e_tag: t.Optional[str] = Field(default=None)
    content_type: t.Optional[str] = Field(default=None)
    user_metadata: t.Optional[dict] = Field(default=None)
    sequencer: t.Optional[str] = Field(default=None)

class MinioS3Data(BaseEventModel):
    """
    MinIO S3 Event Data
    """
    s3_schema_version: t.Optional[str] = Field(default=None)
    configuration_id: t.Optional[str] = Field(default=None)

    bucket: MinioBucket = Field(default_factory=MinioBucket)
    object: MinioObject = Field(default_factory=MinioObject)

class MinioEvent(BaseEventModel):
    """
    MinIO S3 Event Notification
    """
    event_version: t.Optional[str] = Field(default=None)
    event_source: t.Optional[str] = Field(default=None)
    aws_region: t.Optional[str] = Field(default=None)
    event_time: t.Optional[str] = Field(default=None)
    event_name: t.Optional[str] = Field(default=None)
    user_identity: t.Optional[dict] = Field(default=None)
    request_parameters: t.Optional[dict] = Field(default=None)
    response_elements: t.Optional[dict] = Field(default=None)
    s3: t.Optional[MinioS3Data] = Field(default_factory=MinioS3Data)
    source: t.Optional[dict] = Field(default=None)

    @eproperty
    def is_create_event(self) -> bool:
        """
        Returns whether the event is a creation event.
        """
        return fnmatch.fnmatch(self.event_name, "s3:ObjectCreated:*")

    @eproperty
    def is_delete_event(self) -> bool:
        """
        Returns whether the event is a deletion event.
        """
        return fnmatch.fnmatch(self.event_name, "s3:ObjectRemoved:*")
    
    def get_bucket_object(self, scheme: t.Optional[str] = 'mc://', **kwargs) -> 'File':
        """
        Returns the bucket object as a File.
        """
        from lzl.io import File
        return File(path=f"{scheme}{self.s3.bucket.name}")

    def get_file_object(self, scheme: t.Optional[str] = 'mc://', **kwargs) -> 'File':
        """
        Returns a File object representing the MinIO object.
        """
        return self.get_bucket_object(scheme = scheme).joinpath(self.s3.object.key)


class MinioEvents(RootModel):
    root: t.List[MinioEvent]

    def __iter__(self) -> t.Iterator[MinioEvent]:
        return iter(self.root)

    def __getitem__(self, item: int) -> MinioEvent:
        return self.root[item]



from __future__ import annotations

"""
Huggingface Embedding Route

- Supports the usage with `text-embeddings-inference`
"""


from ...embeddings import (
    EmbeddingRoute as BaseEmbeddingRoute,
    EmbeddingObject as BaseEmbeddingObject,
    EmbeddingResponse as BaseEmbeddingResponse,
    logger
)
from lzl.types import validator, eproperty, Field
from lzl.api.openai.types.handlers import ModelContextHandler
from lzl.api.openai.types.base import Usage
# from async_openai.types.context import ModelContextHandler
# from async_openai.types.resources import Usage
from typing import Any, Dict, List, Optional, Union, Set, Type, TYPE_CHECKING


class EmbeddingObject(BaseEmbeddingObject):
    model: Optional[str] = "togethercomputer/m2-bert-80M-32k-retrieval"

    @validator('model', pre=True, always=True)
    def validate_model(cls, v, values: Dict[str, Any]) -> str:
        """
        Validate the model
        """
        if not v:
            if values.get('engine'):
                v = values.get('engine')
            elif values.get('deployment'):
                v = values.get('deployment')
        v = ModelContextHandler.resolve_external_model_name(v)
        return v


class EmbeddingResponse(BaseEmbeddingResponse):

    usage: Optional[Usage] = Field(default_factory = Usage)


    @eproperty
    def consumption(self) -> int:
        """
        Returns the consumption for the completions
        """
        try:
            if not self.usage.prompt_tokens:
                self.usage.prompt_tokens = ModelContextHandler.count_tokens(self.input_object.input, model_name=self.input_object.model)
            return ModelContextHandler.get_consumption_cost(
                model_name = self.input_object.model,
                usage = self.usage,
            )
        except Exception as e:
            logger.error(f"Error getting consumption: {e}")
            return 0


class EmbeddingRoute(BaseEmbeddingRoute):
    input_model: Optional[Type[EmbeddingObject]] = EmbeddingObject
    response_model: Optional[Type[EmbeddingResponse]] = EmbeddingResponse


from __future__ import annotations

"""
FastAPI Helpers
"""

import functools
from lzl.proxied import ProxyObject
from typing import Optional, List, Dict, Any, Union, Type, Callable, TYPE_CHECKING


if TYPE_CHECKING:
    from Cryptodome.Cipher import AES
    from Cryptodome.Cipher._mode_ecb import EcbMode
    from lzl.io import JsonSerializer

serializer: 'JsonSerializer' = ProxyObject(
    obj_getter = 'lzl.io.ser.get_serializer',
    obj_args = ('json',)
)


def create_aes_object(
    secret_key: str,
    access_key: str,
) -> Callable[..., 'EcbMode']:
    """
    Creates an AES Object that can be repeatedly used
    """
    from Cryptodome.Cipher import AES
    return functools.partial(AES.new, secret_key.encode(), AES.MODE_CFB, access_key.encode())




from __future__ import annotations

"""Re-export application environment helpers from ``lzl``."""

# This module is kept for backwards compatibility.

from lzl.types.common import AppEnv, get_app_env

__all__ = ['AppEnv', 'get_app_env']



from __future__ import annotations

import json
import pathlib
import contextlib
from lzl.api import aiohttpx
from lzl.types import BaseSettings, validator, eproperty, Field
from lzo.utils import create_unique_id
from typing import Optional, Dict, Union, Any, Type


class BaseProxySettings(BaseSettings):

    proxy_enabled: Optional[bool] = None
    proxy_endpoint: Optional[str] = None
    proxy_name: Optional[str] = None
    proxy_kind: Optional[str] = None
    proxy_env_name: Optional[str] = None
    proxy_app_name: Optional[str] = None
    proxy_endpoints: Optional[Dict[str, str]] = Field(default_factory = dict)
    proxy_apikeys: Optional[Dict[str, str]] = Field(default_factory = dict)


    @property
    def endpoint(self) -> Optional[str]:
        """
        Returns the Proxy Endpoint
        """
        return self.proxy_endpoint
    
    @property
    def enabled(self) -> Optional[bool]:
        """
        Returns whether the proxy is enabled
        """
        return self.proxy_enabled

    def get_proxy_endpoint(self) -> Optional[str]:
        """
        Returns the proxy endpoint
        """
        if self.proxy_name and self.proxy_endpoints.get(self.proxy_name):
            return self.proxy_endpoints[self.proxy_name]
        for name, endpoint in self.proxy_endpoints.items():
            with contextlib.suppress(Exception):
                resp = aiohttpx.get(endpoint, timeout = 2.0)
                # data = resp.json()
                # if data.get('error'):
                self.proxy_name = name
                return endpoint
        return None

    def init(self, config_path: Optional[pathlib.Path] = None):
        """
        Initializes the core settings
        """
        if config_path: self.load_proxy_config(config_path)
        if self.proxy_endpoint is None: 
            self.proxy_endpoint = self.get_proxy_endpoint()
            self.proxy_enabled = self.proxy_endpoint is not None

    def get_apikey(
        self, source: Optional[str] = None, 
    ) -> str:
        """
        Gets the appropriate API Key for the proxy
        """
        if source:
            source = source.lower()
            for k, v in self.proxy_apikeys.items():
                if k in source: return v
        return self.proxy_apikeys.get('default', None)
    
    def load_proxy_config(
        self,
        path: pathlib.Path,
    ):
        """
        Loads the Proxy Configuration from a File
        """
        if not path.exists(): return
        text = path.read_text()
        if 'env/' in text:
            from lzo.utils import parse_envvars_from_text
            text, _ = parse_envvars_from_text(text)
        
        assert path.suffix in {
            ".yaml", ".yml", ".json"
        }, f"The preset file must be a YAML or JSON file: {path}"
        
        data: Dict[str, Union[Dict[str, str], str]] = {}

        if path.suffix == ".json":
            data = json.loads(text)
        else:
            import yaml
            data = yaml.safe_load(text)
        for k, v in data.items():
            if v is None: continue
            if k in {'endpoint', 'enabled'}: k = f'proxy_{k}'
            if hasattr(self, k): setattr(self, k, v)
            elif hasattr(self, f'proxy_{k}'): setattr(self, f'proxy_{k}', v)
        self.proxy_endpoint = None
        self.proxy_enabled = None

    def update(
        self,
        **kwargs
    ) -> None:
        """
        Updates the Proxy Settings
        """
        for k, v in kwargs.items():
            if v is None: continue
            if k in {'endpoint', 'enabled'}: k = f'proxy_{k}'
            if hasattr(self, k): setattr(self, k, v)
            elif hasattr(self, f'proxy_{k}'): setattr(self, f'proxy_{k}', v)
        self.proxy_endpoint = None
        self.proxy_enabled = None

    def create_proxy_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers for the Proxy Provider
        """
        headers = headers or {}
        return headers

    def configure_proxy_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Configures the Proxy Headers for the Proxy Provider
        """
        return {}
    

    def configure_proxy_headers_for_function(
        self,
        headers: Optional[Dict[str, str]] = None,
        function_name: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Configures the Proxy Headers for the Proxy Provider
        """
        return {}
    
    def configure_proxy_headers_for_json_schema(
        self,
        headers: Optional[Dict[str, str]] = None,
        schema_name: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Configures the Proxy Headers for the Proxy Provider
        """
        return {}
    

    def disable_caching_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Disables the Caching Headers
        """
        return headers
        
    class Config:
        # We use a different prefix here to avoid conflicts
        env_prefix = "OAI_"
        case_sensitive = False


class HeliconeProxySettings(BaseProxySettings):
    """
    The Helicone Proxy Settings
    """
    
    proxy_kind: Optional[str] = 'helicone'


    def create_proxy_headers(
        self,
        name: str,
        config: Dict[str, Any],
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers for Helicone
        """
        headers = {
            'Helicone-OpenAI-Api-Base': config.get('api_base', ''),
            'Helicone-Auth': f"Bearer {self.get_apikey(self.proxy_app_name)}",
            "Helicone-Property-ClientName": name,
            'Content-Type': 'application/json',
        }
        user_id = ''
        if self.proxy_app_name:
            headers['Helicone-Property-AppName'] = self.proxy_app_name
            user_id += self.proxy_app_name
        if self.proxy_env_name:
            headers['Helicone-Property-AppEnv'] = self.proxy_env_name
            if user_id and self.proxy_env_name not in user_id: user_id += f'-{self.proxy_env_name}'
        if user_id: headers['Helicone-User-Id'] = user_id
        if 'properties' in config: properties = config.pop('properties')
        if properties:
            for k, v in properties.items():
                if 'Helicone-Property-' not in k: k = f'Helicone-Property-{k}'
                headers[k] = str(v)
        return headers
    

    def configure_proxy_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers for Helicone
        """
        headers = headers or {}
        return headers
    

    def configure_proxy_headers_for_function(
        self,
        headers: Optional[Dict[str, str]] = None,
        function_name: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Configures the Proxy Headers for the Proxy Provider
        """
        headers = headers or {}
        if function_name: headers['Helicone-Property-FunctionName'] = function_name
        if disable_cache: headers['Helicone-Cache-Seed'] = create_unique_id(length = 5)
        if properties:
            property_meta = {f'Helicone-Property-{k}': str(v) for k, v in properties.items()}
            headers.update(property_meta)
        return headers
    

    def configure_proxy_headers_for_json_schema(
        self,
        headers: Optional[Dict[str, str]] = None,
        schema_name: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Configures the Proxy Headers for the Proxy Provider
        """
        headers = headers or {}
        if schema_name: headers['Helicone-Property-SchemaName'] = schema_name
        if disable_cache: headers['Helicone-Cache-Seed'] = create_unique_id(length = 5)
        if properties:
            property_meta = {f'Helicone-Property-{k}': str(v) for k, v in properties.items()}
            headers.update(property_meta)
        return headers


    def disable_caching_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Disables the Caching Headers
        """
        headers = headers or {}
        headers['Helicone-Cache-Seed'] = create_unique_id(length = 5)
        return headers


ProxyProviderSettings = Union[HeliconeProxySettings, BaseProxySettings]
ProxyProviders = {
    'helicone': HeliconeProxySettings,
}

# This is a proxy settings that passes it through to the downstream provider
class OpenAIProxySettings(BaseSettings):
    """
    The OpenAI Proxy Settings
    """

    proxy_provider: Optional[str] = None

    @eproperty
    def providers(self) -> Dict[str, ProxyProviderSettings]:
        """
        Returns the Proxy Providers
        """
        data = {}
        if self.proxy_provider:
            data[self.proxy_provider] = ProxyProviders[self.proxy_provider]()
        return data
    
    @property
    def current(self) -> Optional[str]:
        """
        Returns the Current Proxy Provider
        """
        return self._extra.get('current')
    
    @current.setter
    def current(self, value: str):
        """
        Sets the Current Proxy Provider
        """
        self.proxy_provider = value
        self._extra['current'] = value
        self.provider = self.get_provider(value)

    @eproperty
    def provider(self) -> Optional[ProxyProviderSettings]:
        """
        Returns the Current Proxy Provider
        """
        return self.providers.get(self.current)


    def register_provider(self, name: str, provider: Union[Type[ProxyProviderSettings], str]):
        """
        Registers a Proxy Provider
        """
        global ProxyProviders
        if isinstance(provider, str):
            from lzl.load import lazy_import
            provider = lazy_import(provider)
        ProxyProviders[name] = provider
    
    def get_provider(self, name: Optional[str] = None) -> ProxyProviderSettings:
        """
        Returns the Proxy Provider
        """
        if name is None: name = self.proxy_provider
        if name not in self.providers: 
            if name not in ProxyProviders: raise ValueError(f"Unsupported Proxy Kind: {name}")
            self.providers[name] = ProxyProviders[name]()
        return self.providers[name]
    
    def create_proxy_headers(
        self,
        name: str,
        config: Dict[str, Any],
        provider: Optional[str] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers
        """
        proxy = self.get_provider(name = provider)
        return proxy.create_proxy_headers(name, config, **properties)
    
    def configure_proxy_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        provider: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers
        """
        headers = headers or {}
        proxy = self.get_provider(name = provider)
        return proxy.configure_proxy_headers(headers, disable_cache = disable_cache, **properties)

    def configure_proxy_headers_for_function(
        self,
        headers: Optional[Dict[str, str]] = None,
        function_name: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        provider: Optional[str] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers
        """
        headers = headers or {}
        proxy = self.get_provider(name = provider)
        return proxy.configure_proxy_headers_for_function(headers, function_name = function_name, disable_cache = disable_cache, **properties)


    def configure_proxy_headers_for_json_schema(
        self,
        headers: Optional[Dict[str, str]] = None,
        schema_name: Optional[str] = None,
        disable_cache: Optional[bool] = None,
        provider: Optional[str] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Creates the Proxy Headers
        """
        headers = headers or {}
        proxy = self.get_provider(name = provider)
        return proxy.configure_proxy_headers_for_json_schema(headers, schema_name = schema_name, disable_cache = disable_cache, **properties)



    def disable_caching_headers(
        self,
        headers: Optional[Dict[str, str]] = None,
        provider: Optional[str] = None,
        **properties: Dict[str, str],
    ) -> Dict[str, Any]:
        """
        Disables the Caching Headers
        """
        headers = headers or {}
        proxy = self.get_provider(name = provider)
        return proxy.disable_caching_headers(headers, **properties)

    def load_config_file(
        self,
        path: pathlib.Path,
    ) -> Dict[str, Union[Dict[str, str], str]]:
        """
        Loads the Proxy Configuration from a File
        """
        if not path.exists(): return
        text = path.read_text()
        if 'env/' in text:
            from lzo.utils import parse_envvars_from_text
            text, _ = parse_envvars_from_text(text)
        
        assert path.suffix in {
            ".yaml", ".yml", ".json"
        }, f"The preset file must be a YAML or JSON file: {path}"
        
        if path.suffix == ".json":
            return json.loads(text)
        import yaml
        return yaml.safe_load(text)

    def load_proxy_config(
        self,
        path: pathlib.Path,
        provider: Optional[str] = None,
    ):
        """
        Loads the Proxy Configuration from a File
        """
        data = self.load_config_file(path = path)
        if not provider: provider = data.pop('proxy_provider')
        if not provider: raise ValueError("Proxy Provider must be provided")
        if provider_class := data.pop('provider_class', None):
            if isinstance(provider_class, str):
                from lzl.load import lazy_import
                provider_class = lazy_import(provider_class)
            self.register_provider(name = provider, provider = provider_class)
        proxy = self.get_provider(name = provider)
        proxy.update(**data)
        
        if not self.proxy_provider: self.current = provider
        # print(f'Proxy Config: {self.proxy_provider}: {self.current} -> {self.provider}')

    def update_proxy_config(
        self,
        provider: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Updates the Proxy Configuration
        """
        if not provider: provider = kwargs.pop('proxy_provider')
        if not provider: raise ValueError("Proxy Provider must be provided")
        proxy = self.get_provider(name = provider)
        proxy.update(**kwargs)
        if not self.proxy_provider: self.current = provider

    def init_proxy(self, provider: Optional[str] = None, config_path: Optional[pathlib.Path] = None):
        """
        Initializes the Proxy
        """
        if provider is None: provider = self.proxy_provider
        if provider is None: raise ValueError("Proxy Provider must be provided")
        self.get_provider(name = provider).init(config_path = config_path)

    @property
    def proxy_app_name(self) -> Optional[str]:
        """
        Returns the Proxy App Name
        """
        return self._extra.get('proxy_app_name', self.provider.proxy_app_name if self.provider else None)
    
    @proxy_app_name.setter
    def proxy_app_name(self, value: str):
        """
        Sets the Proxy App Name
        """
        if not self.provider: raise ValueError("Proxy Provider must be provided")
        self.provider.proxy_app_name = value
        self._extra['proxy_app_name'] = value

    @property
    def proxy_env_name(self) -> Optional[str]:
        """
        Returns the Proxy Env Name
        """
        return self._extra.get('proxy_env_name', self.provider.proxy_env_name if self.provider else None)
    
    @proxy_env_name.setter
    def proxy_env_name(self, value: str):
        """
        Sets the Proxy Env Name
        """
        if not self.provider: raise ValueError("Proxy Provider must be provided")
        self.provider.proxy_env_name = value
        self._extra['proxy_env_name'] = value

    @property
    def enabled(self) -> bool:
        """
        Returns whether the Proxy is Enabled
        """
        return self.provider.enabled if self.current else False

    @property
    def endpoint(self) -> Optional[str]:
        """
        Returns the Proxy Endpoint
        """
        return self.provider.endpoint if self.current else None
        

    class Config:
        # We use a different prefix here to avoid conflicts
        env_prefix = "OAI_"
        case_sensitive = False


from __future__ import annotations

"""
Sqlite Backends
"""

import typing as t
from lzl.logging import logger
from ..base.backend import BaseDatabaseBackend, BackendType

class SqliteBackend(BaseDatabaseBackend):
    """
    Sqlite Backend
    """

    if t.TYPE_CHECKING:
        from .config import SqliteConfig
        config: SqliteConfig

    def index_exists(
        self,
        index_name: str,
    ) -> bool:
        """
        [Sqlite] Checks if the Index Exists
        """
        statement = self.sql.text(
            """
            SELECT EXISTS (
                SELECT FROM sqlite_master WHERE type = 'index' AND name = :index_name
            );
            """
        )
        return self._execute_scalar(statement, {'index_name': index_name})
    
    async def aindex_exists(
        self,
        index_name: str,
    ) -> bool:
        """
        [Sqlite] Checks if the Index Exists
        """
        statement = self.sql.text(
            """
            SELECT EXISTS (
                SELECT FROM sqlite_master WHERE type = 'index' AND name = :index_name
            );
            """
        )
        return await self._aexecute_scalar(statement, {'index_name': index_name})

    def database_exists(
        self,
        db_name: str,
        verbose: t.Optional[bool] = True,
    ) -> bool:
        """
        [Sqlite] Checks if the database exists
        """
        statement = self.sql.text(
            "SELECT 1 FROM sqlite_master WHERE type = 'table' AND name = :db_name;"
        )
        result = self._execute_scalar(statement, {'db_name': db_name})
        if verbose: logger.info(f'Database {db_name} Exists: {result}')
        return result
        
    async def adatabase_exists(
        self,
        db_name: str,
        verbose: t.Optional[bool] = True,
    ) -> bool:
        """
        [Sqlite] Checks if the database exists
        """
        statement = self.sql.text(
            "SELECT 1 FROM sqlite_master WHERE type = 'table' AND name = :db_name;"
        )
        result = await self._aexecute_scalar(statement, {'db_name': db_name})
        if verbose: logger.info(f'Database {db_name} Exists: {result}')
        return result

    def get_table_column_size(
        self,
        table: str,
        column: str,
        schema: t.Optional[str] = None,
        verbose: t.Optional[bool] = False,
    ) -> t.Dict[str, t.Union[str, float]]:
        """
        [Sqlite] Returns the table column size
        """
        pass

    def get_table_column_names(
        self,
        table: str,
        schema: t.Optional[str] = None,
    ) -> t.List[str]:
        """
        [Sqlite] Returns the table column names
        """
        pass


class SQLAlchemyBackend(SqliteBackend):
    """
    SQLAlchemy Backend
    """
    backend: t.Optional[BackendType] = 'sqlalchemy'


class SQLModelBackend(SqliteBackend):
    """
    SQLModel Backend
    """
    backend: t.Optional[BackendType] = 'sqlmodel'

SqliteBackendClasses = t.Union[SQLAlchemyBackend, SQLModelBackend]
SqliteBackendMap = {
    'sqlalchemy': SQLAlchemyBackend,
    'sqlmodel': SQLModelBackend,
}




from __future__ import annotations

"""
Workflow Context Utilities
"""

import os
import psutil
import pathlib
import linecache
import tracemalloc
import contextlib
from pydantic.types import ByteSize
from typing import Any, Dict, List, Optional, TypeVar, Union, TYPE_CHECKING


_cgroup_path = pathlib.Path('/sys/fs/cgroup')
_cgroup_mem_path = pathlib.Path('/sys/fs/cgroup/memory.current')

def display_top(snapshot: tracemalloc.Snapshot, key_type: str = 'lineno', limit: int = 3) -> str:
    """
    Displays the top lines of a tracemalloc snapshot
    """
    snapshot = snapshot.filter_traces((
        tracemalloc.Filter(False, "<frozen importlib._bootstrap>"),
        tracemalloc.Filter(False, "<unknown>"),
    ))
    top_stats = snapshot.statistics(key_type)
    s = f'Top {limit} Lines:\n'

    for index, stat in enumerate(top_stats[:limit], 1):
        frame = stat.traceback[0]
        # replace "/path/to/module/file.py" with "module/file.py"
        filename = os.sep.join(frame.filename.split(os.sep)[-2:])
        fstat = ByteSize(stat.size)
        s += f'- [{index}] {fstat.human_readable(True)} {filename}:{frame.lineno} '
        line = linecache.getline(frame.filename, frame.lineno).strip()
        if line:
            s += f'\t{line}'
        s += '\n'
    other = top_stats[limit:]
    if other:
        size = sum(stat.size for stat in other)
        s += f'- [{len(other)}] {ByteSize(size).human_readable(True)} Other Lines Total\n'
    total = sum(stat.size for stat in top_stats)
    s += f'Total Size: {ByteSize(total).human_readable(True)}'
    return s

def capture_state(
    as_str: bool = False,
    prefix: Optional[str] = None,
    in_k8s: Optional[bool] = None,
) -> Union[str, Dict[str, Any]]:
    """
    Captures the state of the process
    """
    # Memory
    # {'total': 137438953472, 'available': 65029390336, 'percent': 52.7, 'used': 71518945280, 'free': 2075770880, 'active': 62968168448, 'inactive': 61191651328, 'wired': 8550776832}
    state = {
        'memory': psutil.virtual_memory()._asdict(),
        'cpu': psutil.cpu_percent(),
        # 'cpu_count': psutil.cpu_count(),
    }
    if in_k8s:
        with contextlib.suppress(Exception):
            state['memory']['used'] = int(_cgroup_path.joinpath('memory.current').read_text().strip())
            max_mem = _cgroup_path.joinpath('memory.max').read_text().strip()
            if max_mem.isnumeric(): state['memory']['total'] = int(max_mem)
            state['memory']['percent'] = round((state['memory']['used'] / state['memory']['total']) * 100, 1)

    if as_str:
        mem_used = ByteSize(state['memory']['used'])
        mem_total = ByteSize(state['memory']['total'])
        mem_percent = state['memory']['percent']
        prefix = prefix or ''
        return f'{prefix}CPU: {state["cpu"]}%, RAM: {mem_used.human_readable(True)}/{mem_total.human_readable(True)} ({mem_percent}%)'
    return state




"""
The Core File Spec for Cloud Filesystems
"""
import os
from ..path import (
    pathlib,
    PosixPath, 
    WindowsPath, 
    Path, 
    PurePath,
    DEFAULT_ENCODING,
    ON_ERRORS,
    NEWLINE,
    BEGINNING,
    CHUNK_SIZE,
    SEP,
    ENCODING,
    ERRORS,
    FileData,
    FileMode,
    Paths,
    Handle,
)
import typing as t
from .path import (
    CloudFileSystemPath,
    CloudPathT
)
# from .compat
from ..types.base import FilePath, PureFilePosixPath, PureFileWindowsPath
from .paths.aws import (
    FileS3Path,
    FileS3PosixPath,
    FileS3WindowsPath,
    
)
from .paths.minio import (
    FileMinioPath,
    FileMinioPosixPath,
    FileMinioWindowsPath,
)
from .paths.s3c import (
    FileS3CPath,
    FileS3CPosixPath,
    FileS3CWindowsPath,
)
from .paths.r2 import (
    FileR2Path,
    FileR2PosixPath,
    FileR2WindowsPath,
)

PathLike = t.TypeVar(
    'PathLike', 
    bound = t.Union[
        str, 
        os.PathLike[str],
        Path,
        FilePath,
        FileS3Path,
        FileMinioPath,
        FileS3CPath,
        FileR2Path,
    ]
)

ProviderPathLike = t.TypeVar(
    'ProviderPathLike', 
    bound = t.Union[
        FileS3Path,
        FileMinioPath,
        FileS3CPath,
        FileR2Path,
    ]
)

FileLikeT = t.TypeVar(
    'FileLikeT', 
    bound = t.Union[
        Path,
        FilePath,
        FileS3Path,
        FileMinioPath,
        FileS3CPath,
        FileR2Path,
    ]
)

FileLike = t.Union[
    Path,
    FilePath,
    FileS3Path,
    FileMinioPath,
    FileS3CPath,
]

# FileLikeT = t.TypeVar(
#     'FileLikeT',
#     Path,
#     FilePath,
#     FileS3Path,
#     FileMinioPath,
# )


PREFIXES_TO_FP: t.Dict[str, t.Type[ProviderPathLike]] = {
    's3://': FileS3Path,
    'aws://': FileS3Path,

    'mc://': FileMinioPath,
    'mio://': FileMinioPath,
    'minio://': FileMinioPath,

    's3c://': FileS3CPath,
    's3compat://': FileS3CPath,

    'r2://': FileR2Path,
}

_FILESPEC_CLS: t.Tuple[FileLikeT, ...] = (
    FilePath,
    PureFilePosixPath, 
    PureFileWindowsPath,

    FileS3Path,
    FileS3PosixPath,
    FileS3WindowsPath,

    FileMinioPath,
    FileMinioPosixPath,
    FileMinioWindowsPath,

    FileS3CPath,
    FileS3CPosixPath,
    FileS3CWindowsPath,

    FileR2Path,
    FileR2PosixPath,
    FileR2WindowsPath,
)


def as_path(path: PathLike) -> FileLike:
    """
    Given a path-like object, return a path-like object
    
    Create a generic `pathlib.Path`-like abstraction.
    Depending on the input (e.g. `gs://`, `github://`, `ResourcePath`,...), the
    system (Windows, Linux,...), the function will create the right pathlib-like
    abstraction.

    Args:
      path (PathLike): Pathlike object.
    
    Returns:
      A pathlib-like abstraction.
    """
    if isinstance(path, str):
        uri_splits = path.split('://', maxsplit=1)
        if len(uri_splits) > 1:    
            # str is URI (e.g. `gs://`, `github://`,...)
            return PREFIXES_TO_FP[f'{uri_splits[0]}://'](path)
        return FilePath(path)
    elif isinstance(path, _FILESPEC_CLS):
        return path
    elif isinstance(path, os.PathLike):
        return FilePath(path)
    else: raise TypeError(f'Invalid path type: {path!r}')


def get_userhome(as_pathz: bool = True):
    h = os.path.expanduser('~')
    return as_path(h) if as_pathz else h

def get_cwd():
    return os.getcwd()


def resolve_relative(filepath: PathLike) -> str:
    """
    If the filepath is a relative path, convert it to an absolute path
    
    Args:
      filepath (PathLike): The path to the file you want to resolve.
    
    Returns:
      A string.
    """
    if not isinstance(filepath, str): filepath = filepath.as_posix()
    if '://' in filepath: return filepath
    if filepath.startswith('~'): filepath = filepath.replace('~', get_userhome(), 1)
    elif filepath.startswith('../'): filepath = filepath.replace('..', get_cwd(), 1)
    elif filepath.startswith('..'): filepath = filepath.replace('..', f'{pathlib.Path(get_cwd()).parent.parent.as_posix()}/', 1)
    elif filepath.startswith('./'): filepath = filepath.replace('.', get_cwd(), 1)
    elif filepath.startswith('.'): filepath = filepath.replace('.', f'{pathlib.Path(get_cwd()).parent.as_posix()}/', 1)
    return filepath


def get_pathlike(filepath: PathLike, resolve: bool = False) -> FileLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath


def get_filelike(path: PathLike) -> FileLike:
    """
    Transforms the path into a FileLike
    """
    if hasattr(path, 'is_fsspec'): return path
    if hasattr(path, 'as_posix'): return get_pathlike(path.as_posix())
    if isinstance(path, str): return get_pathlike(path)
    if hasattr(path, 'file') and hasattr(getattr(path, 'file'), 'name'): 
        return get_pathlike(path.file.name)
    return get_pathlike(path.name) if hasattr(path, 'name') else path



from __future__ import annotations

import typing as t
from .base import BaseTemporalMixin, MixinKinds, TemporalInputT, TemporalReturnT
from .workflows import TemporalWorkflowMixin
from .activities import TemporalActivityMixin
from .dispatch import TemporalDispatchMixin, ParamT, ReturnT


TemporalMixinT = t.TypeVar('TemporalMixinT', bound = BaseTemporalMixin)
TmpMixinT = t.TypeVar('TmpMixinT')

TemporalWorkflowT = t.TypeVar('TemporalWorkflowT', bound = TemporalWorkflowMixin)
TemporalActivityT = t.TypeVar('TemporalActivityT', bound = TemporalActivityMixin)
TemporalDispatchT = t.TypeVar('TemporalDispatchT', bound = TemporalDispatchMixin)

TemporalObjT = t.Union[TemporalWorkflowT, TemporalActivityT, TemporalDispatchT]


from __future__ import annotations

"""
FileIO Registry - Allows users to register different loaders
"""

import typing as t

if t.TYPE_CHECKING:
    from .main import FileLike

### File Extension to File Loader Mapping
# Maps file extensions to their respective file loaders
# .json -> JSONFileLoader

_ext_to_file_loader: t.Dict[str, t.Union[t.Callable[['FileLike'], None], t.Awaitable['FileLike', None]]] = {}


def register_loader(ext: str, loader: t.Union[t.Callable[['FileLike'], None], t.Awaitable['FileLike', None]], overwrite: t.Optional[bool] = None) -> None:
    """
    Registers a file loader for a specific file extension.
    """
    if not ext.startswith('.'): ext = '.' + ext
    if not overwrite and ext in _ext_to_file_loader:
        raise ValueError(f"Loader for extension '{ext}' is already registered.")
    _ext_to_file_loader[ext] = loader

def get_file_loader(ext: str) -> t.Union[t.Callable[['FileLike'], None], t.Awaitable['FileLike', None]]:
    """
    Returns the file loader for a specific file extension, if registered.
    """
    if not ext.startswith('.'): ext = '.' + ext
    if ext in _ext_to_file_loader:
        return _ext_to_file_loader[ext]
    raise ValueError(f"No loader registered for extension '{ext}'")


from __future__ import annotations

"""
Persistent Dict Based Session Middleware
"""

import os
import typing
from uuid import uuid4
from starlette.datastructures import MutableHeaders, Secret
from starlette.requests import HTTPConnection
from starlette.types import ASGIApp, Message, Receive, Scope, Send
from typing import Any, Dict, List, Optional, Union, Type, TYPE_CHECKING, Callable, Awaitable, Tuple
from lzl.logging import logger, null_logger
from lzl.types import Literal
from .utils import get_pdict, create_session_key

if TYPE_CHECKING:
    from lzl.io import PersistentDict
    from lzl.pool import ThreadPool

class PersistentMiddleware:

    """
    Persistent Middleware
    """

    prehooks: List[Callable[..., Awaitable[None]]] = []
    presend_hooks: List[Callable[..., Awaitable[None]]] = []
    posthooks: List[Callable[..., Awaitable[None]]] = []

    state_prehooks: List[Callable[..., Awaitable[None]]] = []
    state_presend_hooks: List[Callable[..., Awaitable[None]]] = []
    state_posthooks: List[Callable[..., Awaitable[None]]] = []

    def __init__(
        self,
        app: 'ASGIApp',
        secret_key: typing.Union[str, 'Secret'],
        session_cookie: str = "session",
        max_age: typing.Optional[int] = 14 * 24 * 60 * 60,  # 14 days, in seconds
        path: str = "/",
        same_site: Literal["lax", "strict", "none"] = "lax",

        http_only: bool = True,
        https_only: bool = False,

        # Persistent Dict
        redis_url: Optional[str] = None,
        base_key: Optional[str] = None,
        serializer: Optional[str] = 'json',
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        allow_local_fallback: bool = True,
        expiration: Optional[int] = None,
        
        # Extra Configuration
        excluded_session_url_paths: Optional[List[str]] = None,
        inject_app_endpoint: Optional[bool] = True,
        oauth2_enabled: Optional[bool] = False,

        # State Configuration
        state_enabled: Optional[bool] = True,
        state_cookie: Optional[str] = "state",
        state_expiration: Optional[int] = 60 * 5,
        state_kws: Optional[List[str]] = None,

        # Hooks
        prehooks: Optional[List[Callable[..., Awaitable[None]]]] = None,
        presend_hooks: Optional[List[Callable[..., Awaitable[None]]]] = None,
        posthooks: Optional[List[Callable[..., Awaitable[None]]]] = None,

        debug_enabled: Optional[bool] = False,

        **kwargs,
    ) -> None:
        self.app = app
        self.path = path
        self.secret_key = secret_key
        self.same_site = same_site
        self.https_only = https_only
        self.session_cookie = session_cookie
        self.max_age = max_age
        self.security_flags = f"SameSite={same_site}"
        if http_only: self.security_flags += "; HttpOnly"
        if https_only:  # Secure flag can be used with HTTPS only
            self.security_flags += "; Secure"
        self._pdict: Optional['PersistentDict'] = None
        self.oauth2_enabled = oauth2_enabled
        if base_key is None: base_key = session_cookie
        if excluded_session_url_paths is None: excluded_session_url_paths = [
            '/healthz',
            '/health',
            '/metrics',
        ]
        self.excluded_session_url_paths = set(excluded_session_url_paths)
        self.state_cookie = state_cookie
        self.state_enabled = state_enabled
        self.state_expiration = state_expiration
        self.state_kws = state_kws or [
            '_token',
            '_user',
            '_apikey',
        ] if oauth2_enabled else []
        self._app_endpoint_enabled = inject_app_endpoint
        self._pdict_kwargs: Dict[str, Any] = {
            'serializer': serializer,
            'serializer_kwargs': serializer_kwargs,
            'base_key': base_key,
            'redis_url': redis_url or os.getenv('SESSION_REDIS_URL'),
            'allow_local_fallback': allow_local_fallback,
            'expiration': expiration or max_age - 10,
            **kwargs,
        }
        if prehooks: self.prehooks.extend(prehooks)
        if presend_hooks: self.presend_hooks.extend(presend_hooks)
        if posthooks: self.posthooks.extend(posthooks)
        self.logger = logger if debug_enabled else null_logger
        self._extra: Dict[str, Any] = {}
    
    @property
    def pdict(self) -> 'PersistentDict[str, Dict[str, Any]]':
        """
        Returns the Persistent Dict
        """
        if self._pdict is None: self._pdict = get_pdict(**self._pdict_kwargs)
        return self._pdict
    
    @property
    def state_pdict(self) -> 'PersistentDict[str, Dict[str, Any]]':
        """
        Returns the Persistent Dict for State
        """
        if 'state_pdict' not in self._extra:
            self._extra['state_pdict'] = self.pdict.get_child('states', expiration = self.state_expiration)
        return self._extra['state_pdict']
    
    @property
    def state_id_pdict(self) -> 'PersistentDict[str, str]':
        """
        Returns the Persistent Dict for State IDs
        """
        if 'state_id_pdict' not in self._extra:
            self._extra['state_id_pdict'] = self.pdict.get_child('state_id', expiration = 60 * 5)
        return self._extra['state_id_pdict']
    
    @property
    def in_k8s(self) -> bool:
        """
        Returns whether we are in k8s
        """
        if 'in_k8s' not in self._extra: 
            from lzo.utils.system import is_in_kubernetes
            self._extra['in_k8s'] = is_in_kubernetes()
        return self._extra['in_k8s']
    
    @property
    def pooler(self) -> 'ThreadPool':
        """
        Returns the ThreadPool
        """
        if 'pooler' not in self._extra:
            from lzl.pool import ThreadPool
            self._extra['pooler'] = ThreadPool
        return self._extra['pooler']

    def get_app_domain_from_scope(self, scope: 'Scope') -> Optional[str]:
        """
        Returns the domain from the scope
        """
        headers: Dict[str, str] = {
            v[0].decode('utf-8'): v[1].decode('utf-8')
            for v in scope['headers']
        }
        if headers.get(':authority:'):
            return headers[':scheme:'] + '://' + headers[':authority:']
        if headers.get('x-forwarded-host'):
            if headers.get('x-forwaded-scheme'):
                return headers['x-forwaded-scheme'] + '://' + headers['x-forwarded-host']
            if self.in_k8s:
                return 'https://' + headers['x-forwarded-host']
            forwaded_port = headers['x-forwarded-port']
            if '443' in forwaded_port: return 'https://' + headers['x-forwarded-host']
            return 'http://' + headers['x-forwarded-host']
        if headers.get('host'):
            scheme = 'https' if self.in_k8s else 'http'
            return f'{scheme}://' + headers['host']
        return None


    async def get_cookie_header_value(
        self,
        scope: 'Scope',
        session_key: Optional[str] = None, # If no session key is provided, it assumes the session should be deleted
        state_key: Optional[str] = None, # If no session key is provided, it assumes the session should be deleted
        kind: Optional[Literal['session', 'state']] = 'session',
        **kwargs,
    ) -> str:  # sourcery skip: assign-if-exp, reintroduce-else, switch
        """
        Returns the cookie header value
        """
        if kind == 'session':
            if not session_key:
                return "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(  # noqa E501
                    session_cookie=self.session_cookie,
                    data = "null",
                    path = self.path,
                    expires = "expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                    security_flags = self.security_flags,
                )
            return "{session_cookie}={session_key}; path={path}; {max_age}{security_flags}".format(  # noqa E501
                session_cookie = self.session_cookie,
                session_key = session_key,
                path = self.path,
                max_age=f"Max-Age={self.max_age}; " if self.max_age else "",
                security_flags=self.security_flags,
            )
        elif kind == 'state':
            if not state_key:
                return "{state_cookie}={data}; path={path}; {expires}{security_flags}".format(  # noqa E501
                    state_cookie=self.state_cookie,
                    data = "null",
                    path = self.path,
                    expires = "expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
                    security_flags = self.security_flags,
                )
            return "{state_cookie}={state_key}; path={path}; {max_age}{security_flags}".format(  # noqa E501
                state_cookie = self.state_cookie,
                state_key = state_key,
                path = self.path,
                max_age=f"Max-Age={self.max_age}; " if self.max_age else "",
                security_flags=self.security_flags,
            )
        raise ValueError(f'Invalid Kind: {kind}')

    async def run_prehooks(self, scope: 'Scope', connection: 'HTTPConnection', receive: 'Receive', send: 'Send'):
        """
        Runs the prehooks
        """
        if not self.prehooks: return
        for prehook in self.prehooks:
            await self.pooler.asyncish(prehook, self, scope = scope, connection = connection, receive = receive, send = send)
    
    async def run_presend_hooks(self, scope: 'Scope', message: 'Message'):
        """
        Runs the presend hooks
        """
        if not self.presend_hooks: return
        for presend_hook in self.presend_hooks:
            await self.pooler.asyncish(presend_hook, self, scope = scope, message = message)
    
    async def run_posthooks(self, scope: 'Scope', message: 'Message'):
        """
        Runs the posthooks
        """
        if not self.posthooks: return
        for posthook in self.posthooks:
            await self.pooler.asyncish(posthook, self, scope = scope, message = message)
    

    async def run_state_prehooks(self, scope: 'Scope', connection: 'HTTPConnection', receive: 'Receive', send: 'Send'):
        """
        Runs the state prehooks
        """
        if not self.state_prehooks: return
        for prehook in self.state_prehooks:
            await self.pooler.asyncish(prehook, self, scope = scope, connection = connection, receive = receive, send = send)
    
    async def run_state_presend_hooks(self, scope: 'Scope', message: 'Message'):
        """
        Runs the state presend hooks
        """
        if not self.state_presend_hooks: return
        for presend_hook in self.state_presend_hooks:
            await self.pooler.asyncish(presend_hook, self, scope = scope, message = message)
    
    async def run_state_posthooks(self, scope: 'Scope', message: 'Message'):
        """
        Runs the state posthooks
        """
        if not self.state_posthooks: return
        for posthook in self.state_posthooks:
            await self.pooler.asyncish(posthook, self, scope = scope, message = message)

    async def run_mw_hooks(self, stage: Literal['pre', 'presend', 'post'], scope: 'Scope', connection: 'HTTPConnection', receive: 'Receive', send: 'Send'):
        """
        Runs the class middleware hooks.

        This can be used to customize the behavior of a subclassed middleware.
        """
        pass

    async def modify_headers(self, headers: MutableHeaders, scope: 'Scope', **kwargs):
        """
        A hook that can be used to modify the headers
        """
        pass

    async def configure_session(
        self,
        connection: 'HTTPConnection',
        scope: 'Scope',
        **kwargs,
    ) -> Tuple[bool, Optional[str]]:
        """
        Configures the session
        """
        initial_session_was_empty = True
        initial_session_key = None
        if self.oauth2_enabled and (state_id := connection.query_params.get('state')):
            self.logger.info(f'State ID: {state_id}')
            if await self.state_id_pdict.acontains(state_id):
                session_key = await self.state_id_pdict.aget(state_id)
                scope["session"] = await self.pdict.aget(session_key)
                initial_session_was_empty = False
                initial_session_key = session_key
                self.logger.info(f'Found Session Key: {session_key}')
            else: scope["session"] = {}
        elif self.session_cookie in connection.cookies:
            session_key = connection.cookies[self.session_cookie]
            if await self.pdict.acontains(session_key):
                scope["session"] = await self.pdict.aget(session_key)
                initial_session_was_empty = False
                initial_session_key = session_key
            else: scope["session"] = {}
        else: scope["session"] = {}
        return initial_session_was_empty, initial_session_key

    async def save_session(
        self,
        scope: 'Scope',
        message: 'Message',
        session_key: Optional[str],
        initial_session_was_empty: Optional[bool],
        **kwargs,
    ) -> 'MutableHeaders':
        """
        Saves the session
        """
        if scope['session']:
            if scope['session'].get('session_id'):
                if session_key and session_key != scope['session']['session_id']:
                    await self.pdict.aexpire(session_key, 60)
                session_key = scope['session']['session_id']
            
            if not session_key:
                if scope['session'].get('user_id'):
                    session_key = create_session_key(scope['session']['user_id'], self.secret_key)
                else: session_key = create_session_key(str(uuid4()), self.secret_key)
            
            if self.oauth2_enabled and scope['session'].get('state_id'):
                self.logger.info(f'Setting State ID: {scope["session"]["state_id"]} -> Session Key: {session_key}')
                await self.state_id_pdict.aset(scope['session']['state_id'], session_key)
            
            await self.pdict.aset(session_key, scope['session'])
            headers = MutableHeaders(scope=message)
            header_value = await self.get_cookie_header_value(scope, session_key = session_key, kind = 'session')
            headers.append("Set-Cookie", header_value)
        
        elif not initial_session_was_empty:
            # The session has been cleared.
            headers = MutableHeaders(scope=message)
            header_value = await self.get_cookie_header_value(scope, kind = 'session')
            headers.append("Set-Cookie", header_value)
            if session_key: await self.pdict.adelete(session_key)
        return headers

    async def configure_state(
        self,
        connection: 'HTTPConnection',
        scope: 'Scope',
        **kwargs,
    ) -> Tuple[bool, Optional[str]]:  # sourcery skip: use-assigned-variable
        """
        Configures the state
        """
        initial_state_was_empty = True
        initial_state_key = None
        if not self.state_enabled: return initial_state_was_empty, initial_state_key
        if self.oauth2_enabled and (state_id := connection.query_params.get('state')):
            self.logger.info(f'State ID: {state_id}')
            if await self.state_id_pdict.acontains(state_id):
                state_key = await self.state_id_pdict.aget(state_id)
                scope["state"] = await self.state_pdict.aget(state_key)
                initial_state_was_empty = False
                initial_state_key = state_key
                self.logger.info(f'Found State Key: {state_key}')
            else: scope["state"] = {}
        elif self.state_cookie in connection.cookies:
            state_key = connection.cookies[self.state_cookie]
            if await self.state_pdict.acontains(state_key):
                scope["state"] = await self.state_pdict.aget(state_key)
                initial_state_was_empty = False
                initial_state_key = state_key
                self.logger.info(f'Loaded State: {state_key}')
            else: scope["state"] = {}
        else: scope["state"] = {}
        return initial_state_was_empty, initial_state_key

    async def save_state(
        self,
        scope: 'Scope',
        headers: 'MutableHeaders',
        state_key: Optional[str],
        initial_state_was_empty: Optional[bool],
        **kwargs,
    ) -> 'MutableHeaders':
        """
        Saves the State
        """
        if not self.state_enabled: return headers
        if scope['state']:
            if scope['state'].get('state_id'):
                if state_key and state_key != scope['state']['state_id']:
                    await self.state_pdict.aexpire(state_key, 60)
                state_key = scope['state']['state_id']
            if not state_key:
                if scope['state'].get('user_id'):
                    state_key = create_session_key(scope['state']['user_id'], self.secret_key)
                else: state_key = create_session_key(str(uuid4()), self.secret_key)
            
            if self.oauth2_enabled and scope['state'].get('state_id'):
                logger.info(f'Setting State ID: {scope["state"]["state_id"]} -> State Key: {state_key}')
                await self.state_id_pdict.aset(scope['state']['state_id'], state_key)
            
            state = {k:v for k,v in scope['state'].items() if k in self.state_kws}
            self.logger.info(f'Saving State: {state}', prefix = state_key, colored = True)
            await self.state_pdict.aset(state_key, state)
            # await self.state_pdict.aset(state_key, scope['state'])
            header_value = await self.get_cookie_header_value(scope, state_key = state_key, kind = 'state')
            headers.append("Set-Cookie", header_value)
        
        elif not initial_state_was_empty:
            # The state has been cleared.
            header_value = await self.get_cookie_header_value(scope, kind = 'state')
            headers.append("Set-Cookie", header_value)
            if state_key: await self.state_pdict.adelete(state_key)
        return headers
        

    async def __call__(self, scope: 'Scope', receive: 'Receive', send: 'Send') -> None:
        
        if scope["type"] not in ("http", "websocket"):  # pragma: no cover
            await self.app(scope, receive, send)
            return
        
        if scope['path'] in self.excluded_session_url_paths:
            await self.app(scope, receive, send)
            return

        connection = HTTPConnection(scope)
        initial_session_was_empty, initial_session_key = await self.configure_session(
            connection, 
            scope
        )
        initial_state_was_empty, initial_state_key = await self.configure_state(
            connection, 
            scope
        )
        if self._app_endpoint_enabled:
            scope["session"]["app_endpoint"] = self.get_app_domain_from_scope(scope)
        
        await self.run_prehooks(scope, connection, receive, send)
        await self.run_state_prehooks(scope, connection, receive, send)
        await self.run_mw_hooks('pre', scope, connection, receive, send)

        async def send_wrapper(message: 'Message') -> None:
            if message["type"] == "http.response.start":
                await self.run_presend_hooks(scope, message)
                await self.run_state_presend_hooks(scope, message)
                await self.run_mw_hooks('presend', scope, connection, receive, send)
                session_key = initial_session_key
                headers = await self.save_session(
                    scope, 
                    message, 
                    session_key, 
                    initial_session_was_empty
                )
                state_key = initial_state_key
                headers = await self.save_state(
                    scope, 
                    headers, 
                    state_key, 
                    initial_state_was_empty
                )
                await self.modify_headers(headers, scope, session_key = session_key)
            await self.run_posthooks(scope, message)
            await self.run_state_posthooks(scope, message)
            await self.run_mw_hooks('post', scope, connection, receive, send)
            await send(message)
        
        await self.app(scope, receive, send_wrapper)


    @classmethod
    def register_prehook(
        cls,
        prehook: Union[Callable[..., Awaitable[None]], List[Callable[..., Awaitable[None]]]],
        kind: Optional[Literal['session', 'state']] = 'session',
    ):
        """
        Registers a prehook
        """
        if kind == 'state':
            if isinstance(prehook, list): cls.state_prehooks.extend(prehook)
            else: cls.state_prehooks.append(prehook)
        elif kind == 'session':
            if isinstance(prehook, list): cls.prehooks.extend(prehook)
            else: cls.prehooks.append(prehook)

    @classmethod
    def register_presend_hook(
        cls,
        prehook: Union[Callable[..., Awaitable[None]], List[Callable[..., Awaitable[None]]]],
        kind: Optional[Literal['session', 'state']] = 'session',
    ):
        """
        Registers a presend hook
        """
        if kind == 'state':
            if isinstance(prehook, list): cls.state_presend_hooks.extend(prehook)
            else: cls.state_presend_hooks.append(prehook)
        elif kind == 'session':
            if isinstance(prehook, list): cls.prehooks.extend(prehook)
            else: cls.prehooks.append(prehook)


    @classmethod
    def register_posthook(
        cls,
        posthook: Union[Callable[..., Awaitable[None]], List[Callable[..., Awaitable[None]]]],
        kind: Optional[Literal['session', 'state']] = 'session',
    ):
        """
        Registers a posthook
        """
        if kind == 'state':
            if isinstance(posthook, list): cls.state_posthooks.extend(posthook)
            else: cls.state_posthooks.append(posthook)
        elif kind == 'session':
            if isinstance(posthook, list): cls.posthooks.extend(posthook)
            else: cls.posthooks.append(posthook)
        


from __future__ import annotations

"""
Manages the Hatchet State
"""
import abc
# from lazyops.libs.proxyobj import proxied, ProxyObject
# from lazyops.utils import lazy_import
from lzl.proxied import proxied, ProxyObject
from lzl.load import lazy_import
from typing import Optional, Dict, Any, List, Union, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from .client import HatchetClient
    from .session import HatchetSession
    from .worker import Worker
    from .context import Context
    from .workflow.ctx import WorkflowContext
    from .workflow import WorkflowT
    from hatchet_sdk.workflow import WorkflowMeta

    HatchetObjClasses = Type[Union[
        HatchetSession,
        Context,
        Worker,
        WorkflowMeta,
    ]]



# @proxied
class GlobalHatchetContextObject(abc.ABC):
    """
    The Hatchet Context for the Global Hatchet Instance
    """

    ctx: Optional['HatchetSession'] = None
    sessions: Dict[str, 'HatchetSession'] = {}
    classes: Dict[str, 'HatchetObjClasses'] = {}
    workflows: Dict[str, Dict[str, Union[str, 'WorkflowT']]] = {}

    def set_ctx(
        self,
        instance: Optional[str] = None,
        session: Optional['HatchetSession'] = None,
    ):
        """
        Sets the context
        """
        if not instance and not session: raise ValueError('Either instance or session must be provided')
        if instance:
            if instance not in self.sessions: raise ValueError(f'Invalid session instance: {instance}')
            session = self.sessions[instance]
        self.ctx = session

    @property
    def current(self) -> Optional[str]:
        """
        Returns the current session
        """
        return self.ctx.instance if self.ctx else None
    
    @property
    def worker_class(self) -> Type['Worker']:
        """
        Returns the worker class
        """
        return self.classes.get('worker')
    
    @worker_class.setter
    def worker_class(self, value: Type['Worker']):
        """
        Sets the worker class
        """
        self.classes['worker'] = value

    @property
    def context_class(self) -> Type['Context']:
        """
        Returns the context class
        """
        return self.classes.get('context')
    
    @context_class.setter
    def context_class(self, value: Type['Context']):
        """
        Sets the context class
        """
        self.classes['context'] = value

    @property
    def workflow_context_class(self) -> Type['WorkflowContext']:
        """
        Returns the workflow context class
        """
        return self.classes.get('workflow_context')
    
    @workflow_context_class.setter
    def workflow_context_class(self, value: Type['WorkflowContext']):
        """
        Sets the workflow context class
        """
        self.classes['workflow_context'] = value


    @property
    def session_class(self) -> Type['HatchetSession']:
        """
        Returns the session class
        """
        return self.classes.get('session')
    
    @session_class.setter
    def session_class(self, value: Type['HatchetSession']):
        """
        Sets the session class
        """
        self.classes['session'] = value

    @property
    def workflow_meta_class(self) -> Type['WorkflowMeta']:
        """
        Returns the workflow meta class
        """
        return self.classes.get('workflow_meta')
    
    @workflow_meta_class.setter
    def workflow_meta_class(self, value: Type['WorkflowMeta']):
        """
        Sets the workflow meta class
        """
        self.classes['workflow_meta'] = value


    def configure_classes(
        self,
        worker_class: Optional[Union[Type['Worker'], str]] = None,
        context_class: Optional[Union[Type['Context'], str]] = None,
        session_class: Optional[Union[Type['HatchetSession'], str]] = None,
        workflow_meta_class: Optional[Union[Type['WorkflowMeta'], str]] = None,
        workflow_context_class: Optional[Union[Type['WorkflowContext'], str]] = None,
        **kwargs,
    ):
        """
        Configures the global classes that are used to initialize the Hatchet Sessions
        """
        if context_class:
            if isinstance(context_class, str): context_class = lazy_import(context_class)
            self.context_class = context_class
        elif self.context_class is None:
            from .context import Context
            self.context_class = Context
        if session_class:
            if isinstance(session_class, str): session_class = lazy_import(session_class)
            self.session_class = session_class
        elif self.session_class is None:
            from .session import HatchetSession
            self.session_class = HatchetSession
        if worker_class:
            if isinstance(worker_class, str): worker_class = lazy_import(worker_class)
            self.worker_class = worker_class
        elif self.worker_class is None:
            from .worker import Worker
            self.worker_class = Worker
        if workflow_meta_class:
            if isinstance(workflow_meta_class, str): workflow_meta_class = lazy_import(workflow_meta_class)
            self.workflow_meta_class = workflow_meta_class
        elif self.workflow_meta_class is None:
            from hatchet_sdk.workflow import WorkflowMeta
            self.workflow_meta_class = WorkflowMeta
        if workflow_context_class:
            if isinstance(workflow_context_class, str): workflow_context_class = lazy_import(workflow_context_class)
            self.workflow_context_class = workflow_context_class
        elif self.workflow_context_class is None:
            from .workflow.ctx import WorkflowContext
            self.workflow_context_class = WorkflowContext

    def get_workflow(
        self,
        name: str,
        version: Optional[str] = None,
        instance: Optional[str] = None,
    ) -> 'WorkflowT':
        """
        Gets the workflow
        """
        if not instance: instance = self.current or 'default'
        if version: name += f'.{version}'
        if name not in self.workflows.get(instance, {}):
            raise KeyError(f"Invalid Workflow Name: ({instance}) {name} as it is not registered or an importable name")
        if isinstance(self.workflows[instance][name], str):
            self.workflows[instance][name] = lazy_import(self.workflows[instance][name])
        return self.workflows[instance][name]

    def register_workflow(
        self,
        workflow_name: Optional[str] = None,
        workflow_obj: Optional[Union[str, Type['WorkflowT']]] = None,
        workflow_mapping: Optional[Dict[str, Union[str, Type['WorkflowT']]]] = None,
        instance: Optional[str] = None,
        overwrite: Optional[bool] = None,
    ):
        """
        Registers workflows to the global context
        """
        if not instance: instance = self.current or 'default'
        if instance not in self.workflows: self.workflows[instance] = {}
        if workflow_obj:
            if not workflow_name: 
                workflow_name = workflow_obj if isinstance(workflow_obj, str) else workflow_obj.__name__
            if workflow_name in self.workflows[instance] and not overwrite:
                raise ValueError(f"Workflow {workflow_name} already registered for instance {instance}")
            self.workflows[instance][workflow_name] = workflow_obj
        elif workflow_mapping:
            self.workflows[instance].update(workflow_mapping)
        


    def _workflow_register(
        self,
        workflow: 'WorkflowT',
        instance: Optional[str] = None,
    ):
        """
        Registers the workflow

        - this is a private method that is used to register a workflow
        """
        if not instance: instance = self.current or 'default'
        if instance not in self.workflows: self.workflows[instance] = {}
        self.workflows[instance][workflow.workflow_name] = workflow

        
    def gather_workflows(
        self,
        include: Optional[List[str]] = None,
        exclude: Optional[List[str]] = None,
        include_crontasks: Optional[bool] = None,
        only_crontasks: Optional[bool] = None,
        instance: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Type['WorkflowT']]:
        """
        Gathers the workflows
        """
        if not instance: instance = self.current or 'default'
        workflows: Dict[str, 'WorkflowT'] = {}
        include = include or []
        exclude = exclude or []
        if only_crontasks and not include_crontasks: include_crontasks = True
        workflow_items = self.workflows.get(instance, {})
        for workflow_name, ref in workflow_items.items():
            if include and workflow_name not in include: continue
            if exclude and workflow_name in exclude: continue
            wf_ref = self.get_workflow(workflow_name, instance = instance)
            if not include_crontasks and wf_ref.is_crontask: continue
            if only_crontasks and not wf_ref.is_crontask: continue
            workflows[workflow_name] = wf_ref
        return workflows

GlobalHatchetContext: GlobalHatchetContextObject = ProxyObject(
    obj_cls = GlobalHatchetContextObject,
)